"""Includes PreprocessingUserInputService class."""

from typing import Union, List

from organon.fl.mathematics.constants import INT_MAX
from organon.ml.common.enums.target_type import TargetType
from organon.ml.common.helpers.parameter_helper import USE_CLASS_DEFAULT_STR
from organon.ml.common.helpers.user_input_service_helper import get_enum
from organon.ml.preprocessing.services.user_settings.coarse_input_dto import CoarseInputDto
from organon.ml.preprocessing.settings.enums.imputer_type import ImputerType
from organon.ml.preprocessing.settings.enums.scaler_type import ScalerType
from organon.ml.preprocessing.settings.objects.coarse_class_settings import CoarseClassSettings
from organon.ml.preprocessing.settings.objects.imputation_settings import ImputationSettings
from organon.ml.preprocessing.settings.objects.one_hot_encoding_settings import OneHotEncodingSettings
from organon.ml.preprocessing.settings.objects.scaling_settings import ScalingSettings


class PreprocessingUserInputService:
    """Service for validating user input and generating settings for preprocessor services"""

    @classmethod
    def get_coarse_class_settings(cls, cc_input: CoarseInputDto) -> CoarseClassSettings:
        """Validates settings entered by user and generates settings object for CoarseClass service"""
        target_type = get_enum(cc_input.target_column_type, TargetType)
        cc_settings = CoarseClassSettings(cc_input.test_ratio,
                                          cc_input.min_class_size, target_type, cc_input.max_leaf_nodes,
                                          cc_input.stability_check,
                                          cc_input.stability_threshold, cc_input.random_state,
                                          cc_input.positive_class, cc_input.negative_class)
        return cc_settings

    @classmethod
    def get_imputation_settings(cls, numeric_data_method: str,
                                categorical_data_method: str = ImputerType.SIMPLE.name,
                                n_strategy: str = USE_CLASS_DEFAULT_STR, c_strategy: str = 'most_frequent',
                                n_missing_values: Union[int, float, None] = USE_CLASS_DEFAULT_STR,
                                c_missing_values: Union[int, float, str, None] = USE_CLASS_DEFAULT_STR,
                                n_fill_value: Union[float, int] = USE_CLASS_DEFAULT_STR,
                                c_fill_value: Union[
                                    str, float, int] = USE_CLASS_DEFAULT_STR,
                                included_columns: List[str] = None) -> ImputationSettings:
        """Validates settings entered by user and generates settings object for Imputation service"""
        # pylint: disable=too-many-arguments
        numeric_imputer_type = get_enum(numeric_data_method, ImputerType)
        categorical_imputer_type = get_enum(categorical_data_method, ImputerType)
        cls._validate_imputation_settings(numeric_imputer_type, categorical_imputer_type, c_strategy)
        return ImputationSettings(numeric_imputer_type, ImputerType.SIMPLE, n_missing_values,
                                  c_missing_values, n_strategy, c_strategy, n_fill_value, c_fill_value,
                                  included_columns)

    @staticmethod
    def _validate_imputation_settings(numeric_imputer_type: ImputerType,
                                      categorical_imputer_type: ImputerType, c_strategy: str):
        if numeric_imputer_type is None:
            raise ValueError("Numeric imputer type is None.")
        if categorical_imputer_type != ImputerType.SIMPLE:
            raise ValueError(
                "The given categorical data imputer cannot be used. Please select SIMPLE as imputer.")
        if c_strategy not in ["most_frequent", "constant"]:
            raise ValueError(
                "The given categorical column imputation strategy cannot be used. "
                "You can select most_frequent or constant strategies.")

    @classmethod
    def get_scaling_settings(cls, strategy: str) -> ScalingSettings:
        """Validates settings entered by user and generates settings object for Scaling service"""
        strategy_type = get_enum(strategy, ScalerType)
        return ScalingSettings(strategy_type)

    @classmethod
    def get_one_hot_encoding_settings(cls, category_cumulative_ratio: float = 1.0,
                                      category_max_number: int = INT_MAX) -> OneHotEncodingSettings:
        """Validates settings entered by user and generates settings object for OneHotEncoding service"""
        return OneHotEncodingSettings(category_cumulative_ratio, category_max_number)



"""Includes RidgeModeller class"""
from sklearn.linear_model import Ridge

from organon.ml.modelling.algorithms.core.abstractions.base_sklearn_regressor import BaseSklearnRegressor


class RidgeRegressor(BaseSklearnRegressor[Ridge]):
    """Wrapper for sklearn ridge modeller"""

    @property
    def regressor_class(self):
        return Ridge



"""Includes PretrainedModelSettings class"""
from typing import Optional

from organon.ml.common.enums.pretrained_model_type import PretrainedModelType


class PretrainedModelSettings:
    """Settings related to pretrained model in object classification"""

    def __init__(self, model: PretrainedModelType = PretrainedModelType.XCEPTION, weights: Optional[str] = "imagenet"):
        self.model = model
        self.weights = weights



"""
This module includes AfeModelOutput class.
"""

from organon.afe.domain.modelling.businessobjects.afe_feature import AfeFeature
from organon.afe.domain.reporting.base_afe_model_output import BaseAfeModelOutput


class AfeModelOutput(BaseAfeModelOutput[AfeFeature]):
    """AFE output information of model"""



"""
This module includes AfeSupervisedAlgorithmSettings class.
"""
from organon.afe.domain.common.reader_helper import get_values_from_kwargs
from organon.afe.domain.settings.afe_algorithm_settings import AfeAlgorithmSettings
from organon.afe.domain.settings.model_settings import ModelSettings


class AfeSupervisedAlgorithmSettings(AfeAlgorithmSettings):
    """
    Algorithm settings for Automated Feature Extraction with Supervised Learning type
    """

    ATTR_DICT = {
        "training_percentage": float,
        "min_data_in_leaf_and_sample_size_control_ratio": float,
        "model_settings": ModelSettings,
        "final_model_settings": ModelSettings
    }
    ATTR_DICT.update(AfeAlgorithmSettings.ATTR_DICT)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.training_percentage: float = None
        self.min_data_in_leaf_and_sample_size_control_ratio: float = None

        self.model_settings: ModelSettings = None
        self.final_model_settings: ModelSettings = None

        get_values_from_kwargs(self, self.ATTR_DICT, kwargs)



"""Includes DqFileCalculationParameters class."""
from organon.idq.domain.settings.abstractions.dq_base_calculation_parameters import DqBaseCalculationParameters
from organon.idq.domain.settings.input_source.dq_file_input_source_settings import DqFileInputSourceSettings


class DqFileCalculationParameters(DqBaseCalculationParameters[DqFileInputSourceSettings]):
    """DQ calculation parameters for file source type"""



"""todo"""
import math
from typing import List, Dict

from organon.fl.core.businessobjects.idataframe import IDataFrame
from organon.fl.core.collections.sorted_dict import SortedDict
from organon.fl.mathematics.helpers.dataframe_operations_factory import DataFrameOperationsFactory
from organon.fl.mathematics.helpers.idataframe_operations import IDataFrameOperations
from organon.idq.core.dq_constants import DqConstants
from organon.idq.domain.businessobjects.statistics.dq_categorical_statistics import DqCategoricalStatistics
from organon.idq.domain.businessobjects.statistics.dq_sample_numerical_statistics import DqSampleNumericalStatistics
from organon.idq.domain.helpers.statistics.dq_statistical_functions import DqStatisticalFunctions


class DqStatisticsDomainService:
    """todo"""

    @staticmethod
    def compute_categorical_statistics(data_frame: IDataFrame,
                                       missing_values_by_col: Dict[str, List[str]]) \
            -> Dict[str, DqCategoricalStatistics]:
        """Computes nominal statistics for all columns given in missing_values_by_col"""
        df_operations = DataFrameOperationsFactory.get_dataframe_operations(data_frame)

        stats_per_col = {}
        for col, missing_values in missing_values_by_col.items():
            missing_frequencies = df_operations.get_frequencies(data_frame, col, missing_values)
            n_val = df_operations.get_size(data_frame)
            frequencies = df_operations.get_frequencies(data_frame, col)
            stats = DqCategoricalStatistics(missing_values, frequencies)
            stats.n_val = n_val
            stats.n_miss = sum(missing_frequencies.values())
            stats.cardinality = len(frequencies)
            stats_per_col[col] = stats
        return stats_per_col

    @staticmethod
    def compute_numerical_statistics(data_frame: IDataFrame,
                                     missing_values_by_col: Dict[str, List[float]],
                                     tukey_parameter=1.5):
        """Computes numerical statistics for all columns given in missing_values_by_col"""
        df_operations = DataFrameOperationsFactory.get_dataframe_operations(data_frame)

        stats_per_col = {}
        for col, missing_values in missing_values_by_col.items():
            missing_frequencies = df_operations.get_frequencies(data_frame, col, missing_values)
            col_without_missing = df_operations.get_col_df_without_values(data_frame, col, missing_values)
            n_val = df_operations.get_size(col_without_missing)
            if n_val == 0:
                stats_per_col[col] = DqStatisticsDomainService.__get_null_stats(missing_frequencies)
                continue
            stats = DqStatisticsDomainService.__get_stats_for_column(df_operations, col_without_missing, col,
                                                                     missing_frequencies, n_val, tukey_parameter)
            stats_per_col[col] = stats
        return stats_per_col

    @staticmethod
    def __get_stats_for_column(df_operations: IDataFrameOperations, col_without_missing: IDataFrame, col: str,
                               missing_frequencies: Dict[float, int], n_val: int,
                               tukey_parameter: float):
        cardinality = df_operations.get_column_distinct_counts(col_without_missing, col_names=[col])[col]
        compute_freqs = cardinality <= DqConstants.NUMERICAL_STATISTICS_MIN_CARDINALITY

        freqs = {}
        if compute_freqs:
            freqs = df_operations.get_frequencies(col_without_missing, col)
        stats = DqSampleNumericalStatistics()
        stats.missing_values = list(missing_frequencies.keys())
        stats.n_val = n_val
        stats.n_miss = sum(missing_frequencies.values())
        stats.frequencies = SortedDict(freqs)
        stats.compute_frequencies = compute_freqs
        stats.missing_values_frequencies = missing_frequencies
        stats.cardinality = cardinality
        stats.mean = df_operations.get_mean_value(col_without_missing, col, skipna=False)
        stats.min = df_operations.get_min_value(col_without_missing, col, skipna=False)
        stats.max = df_operations.get_max_value(col_without_missing, col, skipna=False)
        stats.sum = df_operations.get_sum_value(col_without_missing, col, skipna=False)

        stats.variance = df_operations.get_variance_value(col_without_missing, col, skipna=False)
        stats.std_dev = math.sqrt(stats.variance)
        stats.percentile_values = SortedDict(df_operations.get_percentile_values(
            col_without_missing, col, DqConstants.DEFAULT_PERCENTILE_LIST))
        stats.quartile_values = SortedDict(df_operations.get_percentile_values(col_without_missing, col,
                                                                               [25, 50, 75]))

        stats.median = stats.quartile_values[50]
        DqStatisticsDomainService.__set_interval_stats_and_trimmed_mean(df_operations, stats, col_without_missing,
                                                                        col, tukey_parameter)
        return stats

    @staticmethod
    def __set_interval_stats_and_trimmed_mean(df_operations: IDataFrameOperations,
                                              stats: DqSampleNumericalStatistics,
                                              col_without_missing: IDataFrame, col: str,
                                              tukey_parameter: float):
        DqStatisticsDomainService.__set_quartile_statistics(stats, tukey_parameter)
        if stats.n_val != 1:
            modified_number_of_intervals = min(DqConstants.NUMERICAL_STATISTICS_MAX_NUM_OF_INTERVALS, stats.n_val)
            frequencies = stats.frequencies
            if frequencies is None or len(frequencies) == 0:
                frequencies = SortedDict(df_operations.get_frequencies(col_without_missing, col))
            stats.interval_statistics = DqStatisticalFunctions.get_balanced_intervals_and_statistics(
                frequencies, modified_number_of_intervals, DqConstants.NUMERICAL_STATISTICS_MIN_INTERVAL_SIZE)
            stats.trimmed_mean = df_operations.get_trimmed_mean_value(col_without_missing, col,
                                                                      stats.tukey_lower_limit,
                                                                      stats.tukey_upper_limit)
        else:
            stats.trimmed_mean = stats.mean

    @staticmethod
    def __get_null_stats(missing_values_frequencies):
        n_miss = sum(missing_values_frequencies.values())
        null_stats = DqSampleNumericalStatistics.get_null_statistics()
        null_stats.missing_values = list(missing_values_frequencies.keys())
        null_stats.frequencies = SortedDict()
        null_stats.missing_values_frequencies = SortedDict(missing_values_frequencies)
        null_stats.n_val = n_miss
        null_stats.n_miss = n_miss
        null_stats.cardinality = 0
        return null_stats

    @staticmethod
    def __set_quartile_statistics(stats: DqSampleNumericalStatistics, tukey_parameter: float):
        quartile_75 = stats.quartile_values[75]
        quartile_25 = stats.quartile_values[25]
        stats.inter_quartile_range = quartile_75 - quartile_25
        delta = tukey_parameter * stats.inter_quartile_range
        stats.tukey_lower_limit = quartile_25 - delta
        stats.tukey_upper_limit = quartile_75 + delta



"""Includes ObjectDetectionTrainer class"""
import os

import yaml

from ultralytics import YOLO

from organon.fl.core.fileoperations import directory_helper
from organon.ml.object_detection.services.enums.map_metric import MapMetrics


class ObjectDetectionTrainingService:
    """Custom YOLOv8 model trainer for object detection"""

    def __init__(self, yaml_data_dir: str, pretrained_model_path: str):
        self.trainer: YOLO = None
        self._yaml_data_dir = yaml_data_dir
        self._pretrained_model_path = pretrained_model_path
        self.is_trained = False

    def train(self, epochs: int, batch_size: int, random_seed: int, output_folder: str, override_data_files: bool):
        """Trains model"""
        if not os.path.exists(self._yaml_data_dir):
            raise ValueError("Directory of yaml file does not exist")
        self.validate_data_dir(self._yaml_data_dir)
        trainer = YOLO(self._pretrained_model_path)
        self.trainer = trainer
        self.trainer.train(data=self._yaml_data_dir, project=output_folder, epochs=epochs, batch=batch_size,
                           seed=random_seed, exist_ok=override_data_files)
        self.is_trained = True

    def evaluate(self, map_metric: MapMetrics):
        """Returns model evaluation metrics for trainer"""
        self._check_trained()

        if map_metric == MapMetrics.MAP50:
            map_score = self.trainer.metrics.box.map50
        elif map_metric == MapMetrics.MAP75:
            map_score = self.trainer.metrics.box.map75
        elif map_metric == MapMetrics.MAP:
            map_score = self.trainer.metrics.box.map
        else:
            raise ValueError("Wrong value for map_metric parameter. Only map50, map75, map values valid.")

        return map_score

    @classmethod
    def validate_data_dir(cls, yaml_data_dir):
        """validate data dir and child folders"""
        if not directory_helper.exists(os.path.join(os.path.dirname(yaml_data_dir), "images")):
            raise ValueError(f"No 'images' directory found in {os.path.dirname(yaml_data_dir)}")
        if not directory_helper.exists(os.path.join(os.path.dirname(yaml_data_dir), "labels")):
            raise ValueError(f"No 'labels' directory found in {os.path.dirname(yaml_data_dir)}")

        original_train_image_dir = cls.get_train_images_dir_path_yaml(yaml_data_dir)
        original_train_labels_dir = cls.get_train_label_dir_path_yaml(yaml_data_dir)
        valid_image_files_dir = cls.get_validation_images_dir_path_yaml(yaml_data_dir)
        valid_labels_dir = cls.get_validation_label_dir_path_yaml(yaml_data_dir)

        cls.validate_image_labels_match(original_train_image_dir, original_train_labels_dir)
        cls.validate_image_labels_match(valid_image_files_dir, valid_labels_dir)

    @classmethod
    def validate_image_labels_match(cls, images_dir: str, labels_dir: str):
        """validation method for images and labels corresponding"""
        all_label_files = [os.path.splitext(ann)[0] for ann in os.listdir(labels_dir)]
        all_img_files = [os.path.splitext(ann)[0] for ann in os.listdir(images_dir)]

        missing_labels = [val for val in all_img_files if val not in all_label_files]
        if missing_labels:
            raise ValueError(
                f"Following images in '{images_dir}' do not have corresponding "
                f"labels: {','.join(missing_labels)}.")

        missing_images = [val for val in all_label_files if val not in all_img_files]
        if missing_images:
            raise ValueError(f"Following labels in '{labels_dir}' "
                             f"do not have corresponding images: {','.join(missing_images)}")

    @classmethod
    def get_validation_images_dir_path_yaml(cls, yaml_file_loc: str):
        """Returns path to validation images folder for given data dir"""
        with open(yaml_file_loc, encoding="utf-8") as file:
            yaml_data = yaml.safe_load(file)
        val_dir = yaml_data.get('val')
        if not directory_helper.exists(val_dir):
            raise ValueError("Images Validation dir does not exists.")
        return val_dir

    @classmethod
    def get_validation_label_dir_path_yaml(cls, yaml_file_loc: str):
        """Returns path to validation images folder for given data dir"""
        with open(yaml_file_loc, encoding="utf-8") as file:
            yaml_data = yaml.safe_load(file)
        val_dir = yaml_data.get('val')
        val_labels_dir = os.path.join(val_dir.replace('images', 'labels'))
        if not directory_helper.exists(val_labels_dir):
            raise ValueError("Labels Validation dir does not exists.")
        return val_labels_dir

    @classmethod
    def get_train_images_dir_path_yaml(cls, yaml_file_loc: str):
        """Returns path to train images folder for given data dir"""
        with open(yaml_file_loc, encoding="utf-8") as file:
            yaml_data = yaml.safe_load(file)
        train_dir = yaml_data.get('train')
        if not directory_helper.exists(train_dir):
            raise ValueError("Images Train dir does not exists.")
        return train_dir

    @classmethod
    def get_train_label_dir_path_yaml(cls, yaml_file_loc: str):
        """Returns path to train images folder for given data dir"""
        with open(yaml_file_loc, encoding="utf-8") as file:
            yaml_data = yaml.safe_load(file)
        train_dir = yaml_data.get('train')
        train_labels_dir = os.path.join(train_dir.replace('images', 'labels'))
        if not directory_helper.exists(train_labels_dir):
            raise ValueError("Labels Train dir does not exists.")
        return train_labels_dir

    @classmethod
    def get_train_images_dir_path(cls, data_dir: str):
        """Returns path to train images folder for given data dir"""
        return os.path.join(data_dir, "images", "train")

    @classmethod
    def get_train_label_dir_path(cls, data_dir: str):
        """Returns path to train annotations folder for given data dir"""
        return os.path.join(data_dir, "labels", "train")

    @classmethod
    def get_validation_images_dir_path(cls, data_dir: str):
        """Returns path to validation images folder for given data dir"""
        return os.path.join(data_dir, "images", "validation")

    @classmethod
    def get_validation_label_dir_path(cls, data_dir: str):
        """Returns path to validation images folder for given data dir"""
        return os.path.join(data_dir, "labels", "validation")

    def _check_trained(self):
        if not self.is_trained:
            raise Exception("Trainer is not trained yet.")



"""Includes BaseDqReportHelperParams class."""
import abc
from typing import List

from organon.idq.domain.businessobjects.dq_comparison_result import DqComparisonResult
from organon.idq.domain.businessobjects.dq_data_column_collection import DqDataColumnCollection
from organon.idq.domain.enums.dq_run_type import DqRunType
from organon.idq.domain.settings.partition_info import PartitionInfo


class BaseDqReportHelperParams(metaclass=abc.ABCMeta):
    """Params for DQ report helper."""

    def __init__(self, data_source_name: str, comparison_results: List[DqComparisonResult],
                 data_column_collection: DqDataColumnCollection, run_type: DqRunType,
                 partition: List[PartitionInfo] = None):
        self.data_source_name = data_source_name
        self.partition: List[PartitionInfo] = partition
        self.comparison_results = comparison_results
        self.data_column_collection = data_column_collection
        self.run_type = run_type
        self.filter_str: str = None



"""Includes DatetimeToken enum class."""
from enum import Enum


class DatetimeToken(Enum):
    """Datetime tokens"""
    NOT_AVAILABLE = 0
    MILLISECOND = 1
    SECOND = 2
    MINUTE = 3
    HOUR = 4
    DAY = 5
    MONTH = 6
    YEAR = 7



"""
This file creates extensions for pyx files.
"""
import platform

import numpy
from Cython.Build import cythonize
from setuptools import setup, Extension

os_type_name: str = platform.system()

if "WINDOWS" in os_type_name.upper():
    omp_extra_compile_arg = "/openmp"
    omp_extra_link_arg = "/openmp"
else:
    omp_extra_compile_arg = "-fopenmp"
    omp_extra_link_arg = "-fopenmp"

ext_modules = [
    Extension(
        name='organon.afe.domain.modelling.helper.data_frame_builder_helper',
        sources=[r'./organon/afe/domain/modelling/helper/data_frame_builder_helper.pyx'],
        extra_compile_args=[omp_extra_compile_arg],
        extra_link_args=[omp_extra_link_arg],
        include_dirs=[numpy.get_include()],
        language='c++'
    ),
    Extension(
        name='organon.fl.mathematics.sweep.pyx_files.sweep_helper',
        sources=[r'./organon/fl/mathematics/sweep/pyx_files/sweep_helper.pyx'],
        extra_compile_args=[omp_extra_compile_arg],
        extra_link_args=[omp_extra_link_arg],
        include_dirs=[numpy.get_include()],
        language='c++'
    ),
]

if __name__ == "__main__":
    setup(name="CythonFiles", ext_modules=cythonize(ext_modules))



"""Includes math constants"""

import numpy as np

DOUBLE_MIN = np.finfo(np.float64).min
DOUBLE_MAX = np.finfo(np.float64).max
INT_MAX = np.iinfo(np.int64).max
INT_MIN = np.iinfo(np.int64).min
NEGATIVE_INFINITY = np.NINF
POSITIVE_INFINITY = np.PINF



"""Includes TfDatasetDataSourceHandler class"""
from typing import TYPE_CHECKING

from pandas import DataFrame

from organon.ml.object_classification.common.object_classification_constants import \
    ObjectClassificationConstants
from organon.ml.object_classification.domain.data_sources.base_data_source_handler import \
    BaseDataSourceHandler, _DataSourceDataTypeT
from organon.ml.object_classification.domain.objects.data_source_settings import DataSourceSettings

if TYPE_CHECKING:
    import tensorflow as tf  # noqa


class TfDatasetDataSourceHandler(BaseDataSourceHandler["tf.data.Dataset"]):
    """Class for handling tensorflow dataset input"""

    def __init__(self, data: _DataSourceDataTypeT, settings: DataSourceSettings):
        super().__init__(data, settings)
        self._dataset = self.data

    def process_data(self):
        self._transform_tf_dataset()

    def initialize_probs_df(self) -> DataFrame:
        return DataFrame({ObjectClassificationConstants.LABELS_COL_NAME: self.get_labels()})

    def _get_binary_labels(self):
        labels = self._extract_labels()
        return labels

    def _get_multiclass_labels(self):
        return self._extract_labels()



"""Includes IDQ class."""
from typing import List

from organon.idq.core.dq_constants import DqConstants
from organon.idq.domain.services.user_input_service import UserInputService
from organon.idq.services.base_idq_executor import BaseIDQ
from organon.idq.services.user_settings.dq_user_input import DqUserInput


class IDQ(BaseIDQ[UserInputService, DqUserInput]):
    """User service class for IDQ."""

    def _get_user_settings_instance(self) -> DqUserInput:
        return DqUserInput()

    def _get_user_input_service(self) -> UserInputService:
        return UserInputService(self._user_settings)

    def set_calculation_settings(self,
                                 input_source_settings: dict,
                                 max_nominal_cardinality_count: int = DqConstants.MAX_NOMINAL_CARDINALITY_COUNT_DEFAULT,
                                 outlier_parameter: float = DqConstants.OUTLIER_PARAM_DEFAULT,
                                 name: str = None,
                                 partitions: List[List[dict]] = None):
        """
        Sets calculation settings

        :param input_source_settings:
        :param max_nominal_cardinality_count:
        :param outlier_parameter:
        :param name:
        :param partitions:
        """
        # todo parametre açıklamaları eklenecek
        # pylint: disable=too-many-arguments
        self._user_settings.set_multi_partition_calculation(input_source_settings, max_nominal_cardinality_count,
                                                            outlier_parameter, None,
                                                            name, partitions=partitions)



"""
This class keeps the helper functions that are used in gaining information about the environment.
"""
import getpass
import logging
import os
import platform
import signal
import socket
from typing import List

import psutil

from organon.fl.core.helpers import string_helper
from organon.fl.generic.datadtos.environment_info_dto import EnvironmentInfoDto
from organon.fl.generic.datadtos.process_info_dto import ProcessInfoDto
from organon.fl.generic.enums.os_type_id import OsType


def get_environment_info_dto() -> EnvironmentInfoDto:
    """
    Returns the information about environment.
    :return: Information about environment
    """
    dto: EnvironmentInfoDto = EnvironmentInfoDto(get_machine_name(), get_user_name())
    return dto


def get_machine_name() -> str:
    """
    Returns the machine name.
    :return: machine name
    """
    host_name: str = ""
    try:
        host_name = socket.gethostname()
    except (ValueError, TypeError) as ex:
        logging.error("Hostname can not be resolved")
        logging.error(ex)
    return host_name


def get_ipv4_address() -> str:
    """
    Returns the ipv4 address.
    :return: ipv4 address
    """
    ip_address: str = ""
    try:
        host_name = socket.gethostname()
        ip_address = socket.gethostbyname(host_name)
    except (ValueError, TypeError, socket.gaierror) as ex:
        logging.error("IpAddress can not be resolved")
        logging.error(ex)
    return ip_address


def get_user_name() -> str:
    """
    Returns user name.
    :return: user name
    """
    user_name: str = getpass.getuser()
    return user_name


def get_process_by_process_id(process_id: int) -> ProcessInfoDto:
    """
    Returns process according to the process id.
    :param process_id: id of the process
    :return: information about Process
    """
    current_process_list: list = get_process_list("")
    for dto in current_process_list:
        if dto.process_id == process_id:
            return dto
    return None


def get_process_list(process_name_equals: str = None) -> List[ProcessInfoDto]:
    """
    Returns a list of ProcessInfoDto objects
    :param process_name_equals: process name
    :return: List of ProcessInfoDto objects
    """
    if process_name_equals is None:
        return get_process_list("")

    proc_list: List[ProcessInfoDto] = []
    try:
        for proc in psutil.process_iter():
            process_name = proc.name()
            process_id = proc.pid
            if string_helper.is_null_or_empty(process_name_equals) or process_name_equals == process_name:
                process = ProcessInfoDto(process_name, process_id)
                proc_list.append(process)
    except (KeyError, TypeError, ValueError) as ex:
        logging.error("An error occurred while accessing the process.")
        logging.error(ex)
    return proc_list


def get_total_cores() -> int:
    """
    Returns number of total cores of the computer
    :return: number of total cores
    """
    return int(os.cpu_count())


def get_total_memory_in_bytes() -> float:
    """
    Returns total memory bytes.
    :return: total memory bytes
    """
    total_memory_bytes: int = psutil.virtual_memory().total
    return total_memory_bytes


def get_running_process_id() -> str:
    """
    Returns running process id
    :return: running process id
    """
    return str(os.getpid())


def kill_process_by_id(process_id: int):
    """
    Kills process by process id.
    :param process_id: ID of the process
    :return: nothing
    """
    os.kill(process_id, signal.SIGILL)


def get_os_type() -> OsType:
    """
    Returns operating system type id
    :return: operating system type id
    """
    os_type_name: str = platform.system()
    if "WINDOWS" in string_helper.to_upper_eng(os_type_name):
        return OsType.WINDOWS
    if "LINUX" in string_helper.to_upper_eng(os_type_name):
        return OsType.LINUX
    return OsType.NOT_AVAILABLE


def get_working_dir() -> str:
    """
    returns the path of the current working directory
    :return: path of the current working directory
    """
    work_dir = os.getcwd()
    return work_dir



"""Includes ObjectDetector class"""
import os
from typing import List, Dict

from ultralytics import YOLO

from organon.fl.logging.helpers.log_helper import LogHelper
from organon.ml.object_detection.services.object_detection_trainer import ObjectDetectionTrainer
from organon.ml.object_detection.services.objects.detection_info import DetectionInfo


class ObjectDetector:
    """Class for detecting objects in images using a trained model"""

    def __init__(self, pretrained_model_path: str):
        """
        Initializes ObjectDetector to use a pretrained model for detection

        """
        self.model_path = None
        self._detector = None
        self.pretrained_model_path = None

        if pretrained_model_path != -1:
            self._validate_pretrained_model_path(pretrained_model_path)
            self.pretrained_model_path = pretrained_model_path

            try:
                self._detector = YOLO(pretrained_model_path)
            except Exception as exc:
                msg = "Error occurred while loading pretrained model. " \
                      "Make sure the model in given path is a pretrained model." \
                      "If you want to run detection with a model you trained using ObjectDetectionTrainer," \
                      "use from_custom_model classmethod to initialize the detector."
                LogHelper.error(msg)
                raise Exception(msg) from exc

    def detect(self, input_image_path: str, output_folder: str = None, objects_to_detect: List[str] = None,
               save: bool = True, save_txt: bool = True) -> List[DetectionInfo]:
        """
        Detect object in input image.
        :param input_image_path: Path to image to detect objects from
        :param objects_to_detect: List of objects to detect in image. Cannot be used for detection with custom models.
            To see all available objects for detection call 'get_all_object_names' method.
        """
        if output_folder is not None:
            if not os.path.exists(output_folder):
                raise ValueError("Output Folder path does not exists.")

        results = self._get_detections(input_image_path, objects_to_detect=objects_to_detect, save=save,
                                       save_txt=save_txt, output_folder=output_folder)
        detection_info = []
        if results:
            for detection in results:
                detection_info.append(DetectionInfo(object_name=detection.boxes.cls, confidence=detection.boxes.conf,
                                                    box_points=detection.boxes.xyxy))
        return detection_info

    def _get_detections(self, input_image_path: str, output_folder: str = None,
                        objects_to_detect: List[str] = None, save: bool = True,
                        save_txt: bool = True):

        if self.model_path:
            if objects_to_detect:
                raise NotImplementedError("objects_to_detect feature is not implemented for detection with "
                                          "custom trained models")
        try:
            if objects_to_detect:
                object_key_list = self._validate_objects_to_detect(objects_to_detect)

                if output_folder is not None:
                    results = self._detector(input_image_path,
                                             save=save, save_txt=save_txt, classes=object_key_list,
                                             project=output_folder)
                else:
                    results = self._detector(input_image_path,
                                             save=save, save_txt=save_txt, classes=object_key_list,
                                             project=os.path.dirname(input_image_path))
            else:
                if output_folder is not None:
                    results = self._detector(input_image_path, save=save, save_txt=save_txt, project=output_folder)
                else:
                    results = self._detector(input_image_path, save=save, save_txt=save_txt,
                                             project=os.path.dirname(input_image_path))
        except Exception as exc:
            raise exc

        return results

    def get_all_object_names(self) -> Dict:
        """Returns all object available for detection"""
        return self._detector.names

    @classmethod
    def from_custom_model(cls, model_path: str):
        """
        Initializes ObjectDetector to use a custom trained model for detection

        :param str model_path: Path of a model trained with ObjectDetectionTrainer. Should be given if you will not use
            the pretrained model.
        """
        detector = ObjectDetector(pretrained_model_path=-1)
        detector.model_path = model_path
        detector._detector = YOLO(model_path)
        return detector

    @classmethod
    def load_pretrained_model_from_web(cls, target_path: str,
                                       model_type: str = "small"):
        """Loads a pretrained model to given file path after downloading from web. The default model to be downloaded is
        'YOLOv8s' and its approx 25 MB
        :param target_path: File path to download the pretrained model in
        :param model_type: Model type keyword to be downloaded
        """
        ObjectDetectionTrainer.load_pretrained_model_from_web(target_path, model_type=model_type)

    @classmethod
    def _validate_pretrained_model_path(cls, pretrained_model_path: str):
        if pretrained_model_path is None:
            raise ValueError("Please enter pretrained_model_path. You can use 'load_pretrained_model_from_web'"
                             " method to load a model from internet to a file.")
        if not os.path.isfile(pretrained_model_path):
            raise ValueError(f"No model file found in {pretrained_model_path}.")

    def _validate_objects_to_detect(self, object_names: List[str]):
        all_obj_names = self.get_all_object_names()
        non_existing_objects = [val for val in object_names if val not in all_obj_names.values()]
        if non_existing_objects:
            raise ValueError(f"Following objects cannot be detected by model: {','.join(non_existing_objects)}."
                             " Call get_all_object_names to see valid object names")
        object_key_list = [key for key, value in all_obj_names.items() if value in object_names]

        return object_key_list



"""Includes helper functions for target-trx source reading"""
import pandas as pd
from pandas.core.dtypes.common import is_string_dtype, is_integer_dtype


def validate_entity_column(source_name: str, data_frame: pd.DataFrame, entity_col_name: str):
    """Validates entity column - used for both trx and target frame"""
    if not is_string_dtype(data_frame.dtypes[entity_col_name]) and not is_integer_dtype(
            data_frame.dtypes[entity_col_name]):
        raise ValueError(f"{source_name} data entity column type should be either integer or string")



"""Includes BaseReporterService class"""
import abc
from typing import TypeVar, Generic

import pandas as pd

from organon.ml.reporting.domain.objects.base_report import BaseReport
from organon.ml.reporting.settings.objects.reporter_settings import ReporterSettings

ReportType = TypeVar("ReportType", bound=BaseReport)


class BaseReporterService(Generic[ReportType], metaclass=abc.ABCMeta):
    """Base class for reporter services"""

    @classmethod
    @abc.abstractmethod
    def execute(cls, settings: ReporterSettings) -> ReportType:
        """Executes reporting and returns report"""

    @classmethod
    def _get_row_count(cls, data: pd.DataFrame):
        return len(data)



"""Includes BaseFeatureSelectionService class."""
import abc
from typing import TypeVar, Generic, Dict, List

import numpy as np
import pandas as pd

from organon.fl.core.helpers.data_frame_helper import get_numerical_column_names
from organon.fl.logging.helpers.log_helper import LogHelper
from organon.ml.feature_selection.domain.objects.base_selection_output import BaseSelectionOutput
from organon.ml.feature_selection.domain.objects.settings.base_feature_selection_settings import \
    BaseFeatureSelectionSettings
from organon.fl.modelling.unsupervised_feature_extractor import UnsupervisedFeatureExtractor
from organon.ml.preprocessing.services.preprocessor import Preprocessor
from organon.ml.preprocessing.settings.enums.imputer_type import ImputerType

FeatureSelectionSettingsType = TypeVar("FeatureSelectionSettingsType", bound=BaseFeatureSelectionSettings)
FeatureSelectionOutputType = TypeVar("FeatureSelectionOutputType", bound=BaseSelectionOutput)


class SelectionRunSettings:
    """All settings to run selection after preprocessing"""

    def __init__(self, preprocessed_data: pd.DataFrame, settings,
                 ohe_columns: Dict[str, List[str]]):
        self.data = preprocessed_data
        self.settings = settings
        self.ohe_columns = ohe_columns


class BaseFeatureSelectionService(Generic[FeatureSelectionSettingsType, FeatureSelectionOutputType],
                                  metaclass=abc.ABCMeta):
    """Base service class for feature selection services"""

    def __init__(self):
        self.settings: FeatureSelectionSettingsType = None

    @classmethod
    def run_selection(cls, settings: FeatureSelectionSettingsType) -> FeatureSelectionOutputType:
        """Run selection algorithm and return output with selected features"""
        cls._validate_settings(settings)
        data, ohe_columns = cls._preprocess(settings.data)
        run_settings = SelectionRunSettings(data, settings, ohe_columns)
        output = cls._run(run_settings)
        output.selected_features = cls._get_normalized_selected_features(output.selected_features, ohe_columns)
        return output

    @classmethod
    @abc.abstractmethod
    def _run(cls, run_settings: SelectionRunSettings) -> BaseSelectionOutput:
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def _validate_settings(cls, settings: FeatureSelectionSettingsType):
        raise NotImplementedError

    @classmethod
    def _preprocess(cls, data: pd.DataFrame):
        numerical_columns = get_numerical_column_names(data)
        columns_with_nan_values = [col for col in numerical_columns if data[col].isnull().values.any()]
        execute_imputation = len(columns_with_nan_values) > 0
        execute_ohe = len(numerical_columns) != len(data.columns)
        ohe_columns_dict = None
        if execute_imputation or execute_ohe:
            data = data.copy()
            if execute_imputation:
                imputation_service = Preprocessor().get_imputation_service(numeric_data_method=ImputerType.SIMPLE.name,
                                                                           n_strategy="median",
                                                                           included_columns=numerical_columns)
                imputation_service.fit(data)
                data = imputation_service.transform(data)
            if execute_ohe:
                ohe_service = Preprocessor.get_one_hot_encoding_service()
                ohe_service.fit(data)
                data = ohe_service.transform(data)
                ohe_columns_dict = ohe_service.columns_dict
        return data, ohe_columns_dict

    @classmethod
    def _get_normalized_selected_features(cls, selected_features: List[str], ohe_columns: Dict[str, List[str]]):
        if ohe_columns is None:
            return selected_features.copy()
        reverse_dict = {}
        for str_col, generated_num_cols in ohe_columns.items():
            for col in generated_num_cols:
                reverse_dict[col] = str_col
        actual_features = set()
        for feature in selected_features:
            if feature in reverse_dict:
                actual_col = reverse_dict[feature]
                actual_features.add(actual_col)
            else:
                actual_features.add(feature)
        return list(actual_features)

    @classmethod
    def _get_features_with_sweep(cls, data: pd.DataFrame, sweep_args: dict):
        sweep_args = {} if sweep_args is None else sweep_args.copy()
        n_threads = 1
        if "n_threads" in sweep_args:
            n_threads = sweep_args.pop("n_threads")
        if "max_col_count" in sweep_args:
            max_col_count = sweep_args.pop("max_col_count")
            sweep_args["max_iter"] = max_col_count
        df_as_dict = cls.__get_df_as_dict_for_sweep(data)
        extractor = UnsupervisedFeatureExtractor(df_as_dict, **sweep_args)
        return extractor.execute(num_threads=n_threads)

    @classmethod
    def __get_df_as_dict_for_sweep(cls, data: pd.DataFrame):
        df_dict = {}
        for col in data.columns:
            if data.dtypes[col].type != np.float32:
                df_dict[col] = data[col].astype(np.float32)
            else:
                df_dict[col] = data[col]
        return df_dict

    @classmethod
    def _validate_original_data_for_sweep(cls, data: pd.DataFrame):
        invalid_cols = []
        for col in get_numerical_column_names(data):
            if data.dtypes[col].type != np.float32:
                invalid_cols.append(col)
        if invalid_cols:
            LogHelper.warning(
                "dtypes of these numeric columns will be converted to np.float32 before sweep: " +
                ",".join(invalid_cols))



"""todo"""
from typing import List

from organon.idq.domain.businessobjects.dq_calculation_result import DqCalculationResult
from organon.idq.domain.businessobjects.dq_comparison_result import DqComparisonResult
from organon.idq.domain.businessobjects.dq_control_parameters import DqControlParameters
from organon.idq.domain.businessobjects.main_sample_results import MainSampleResults
from organon.idq.domain.controls.base_dq_control import BaseDqControl
from organon.idq.domain.controls.column_mean_control import ColumnMeanControl
from organon.idq.domain.controls.data_source_stats_time_series_control import DataSourceStatsTimeSeriesControl
from organon.idq.domain.controls.data_source_stats_traffic_light_control import DataSourceStatsTrafficLightControl
from organon.idq.domain.controls.duplicate_keys_control import DuplicateKeysControl
from organon.idq.domain.controls.empty_table_control import EmptyTableControl
from organon.idq.domain.controls.psi_nominal_signal_control import PsiNominalSignalControl
from organon.idq.domain.controls.psi_numeric_signal_control import PsiNumericSignalControl
from organon.idq.domain.controls.stable_column_control import StableColumnControl
from organon.idq.domain.controls.table_columns_control import TableColumnsControl
from organon.idq.domain.controls.unexpected_nominal_values_control import UnexpectedNominalValuesControl
from organon.idq.domain.controls.unexpected_numerical_values_control import UnexpectedNumericalValuesControl
from organon.idq.domain.enums.dq_control_type import DqControlType
from organon.idq.domain.settings.abstractions.dq_base_calculation_parameters import DqBaseCalculationParameters
from organon.idq.domain.settings.abstractions.dq_base_comparison_parameters import DqBaseComparisonParameters


class DqComparisonTaskExecutor:
    """todo"""

    def __init__(self, comparison_parameters: DqBaseComparisonParameters,
                 calculation_parameters: List[DqBaseCalculationParameters],
                 calculation_results: List[DqCalculationResult],
                 control_types_to_execute: List[DqControlType] = None):
        self.dq_control_params: DqControlParameters = self.__initialize_params(comparison_parameters,
                                                                               calculation_parameters,
                                                                               calculation_results)
        self.controls_to_execute = self.__get_control_classes_to_execute(control_types_to_execute)

    @classmethod
    def get_control_classes(cls):
        """Returns all control classes for executor"""
        return [TableColumnsControl, EmptyTableControl, StableColumnControl, UnexpectedNumericalValuesControl,
                UnexpectedNominalValuesControl, DuplicateKeysControl, ColumnMeanControl,
                DataSourceStatsTrafficLightControl, DataSourceStatsTimeSeriesControl, PsiNominalSignalControl,
                PsiNumericSignalControl]

    @classmethod
    def __get_control_classes_to_execute(cls, control_types: List[DqControlType]):
        if control_types is None:
            return cls.get_control_classes().copy()
        return [control_class for control_class in cls.get_control_classes()
                if control_class.get_control_type() in control_types]

    @staticmethod
    def __initialize_params(comparison_parameters: DqBaseComparisonParameters,
                            calculation_parameters: List[DqBaseCalculationParameters],
                            calculation_results: List[DqCalculationResult]):

        control_params = DqControlParameters()
        control_params.comparison_parameters = comparison_parameters
        if len(calculation_results) == 0:
            raise ValueError("Calculations cannot be empty")
        if len(calculation_parameters) != len(calculation_results):
            raise ValueError("Calculation parameters and results lists should be same length")
        control_params.control_results = calculation_results[:-1]
        control_params.control_parameters = calculation_parameters[:-1]
        control_params.test_calc_parameters = calculation_parameters[-1]
        control_params.test_calc_result = calculation_results[-1]
        control_params.test_group_info = {group.group_type: group for group in comparison_parameters.test_groups}
        return control_params

    def execute(self) -> MainSampleResults:
        """todo"""
        results = self.execute_sequentially()
        output = MainSampleResults()
        output.comparison_results = results
        return output

    def execute_sequentially(self):
        """todo"""
        results = []
        for control_class in self.controls_to_execute:
            control = control_class(self.dq_control_params)
            results.extend(self.__execute_control(control))
        return results

    def __execute_control(self, control: BaseDqControl) -> List[DqComparisonResult]:
        test_group_info = self.dq_control_params.test_group_info[control.get_test_group_type()]
        if test_group_info.inclusion_flag is False:
            return []
        return control.execute()



"""
This module includes helper functions for pandas DataFrames.
"""
from typing import List, Dict

import numpy as np
import pandas as pd
from pandas.core.dtypes.common import is_numeric_dtype, is_string_dtype, is_datetime64_any_dtype

from organon.fl.core.enums.column_native_type import ColumnNativeType
from organon.fl.core.executionutil import parallel_execution_helper
from organon.fl.core.helpers.object_helper import get_attribute_dict


def objects_to_data_frame(objects: List[object], column_name_map: Dict[str, str] = None) -> pd.DataFrame:
    """
    Converts list of objects to DataFrame
    :param column_name_map: dictionary to map attribute names to dataframe column names
    :param objects: List of objects
    """
    if column_name_map is None:
        column_name_map = {i: i for i in list(get_attribute_dict(objects[0]).keys())}
        # columns = list(objects[0].__dict__.keys())
    data_frame = pd.DataFrame(columns=list(column_name_map.values()))
    for obj in objects:
        obj_dict = get_attribute_dict(obj)
        values_dict = {column_name_map[attr]: obj_dict[attr] for attr in column_name_map}
        data_frame = data_frame.append(values_dict, ignore_index=True)
    return data_frame


def dataframe_to_objects(data_frame: pd.DataFrame, cls, column_name_map: Dict[str, str] = None):
    """
    Converts the given data frame to list of objects of given class.
    :param data_frame:
    :param cls: Class of objects
    :param column_name_map: map of attribute names to dataframe column names
    :return: List of objects of given class
    """
    if column_name_map is None:
        column_name_map = {i: i for i in list(data_frame.columns)}

    entities = []
    for rowiter in data_frame.iterrows():
        row = rowiter[1]
        entity = cls()
        for attr in get_attribute_dict(entity):
            df_column_name = attr
            if attr in column_name_map:
                df_column_name = column_name_map[attr]
            if df_column_name in row:
                val = row[df_column_name]
                if isinstance(val, pd.Timestamp):
                    val = val.to_pydatetime()
                elif isinstance(val, float):
                    if np.isnan(val):
                        val = None
                setattr(entity, attr, val)
        entities.append(entity)
    return entities


def cast_to_datetime(data_frame: pd.DataFrame, column_name: str):
    """
    Casts given column of dataframe to datetime.
    """
    data_frame[column_name] = pd.to_datetime(data_frame[column_name])
    return data_frame


def cast_to_int(data_frame: pd.DataFrame, column_name: str):
    """
    Casts given column of dataframe to int.
    """
    return data_frame.astype({column_name: int})


def cast_to_long(data_frame: pd.DataFrame, column_name: str):
    """
    Casts given column of dataframe to long(np.int64).
    """
    return data_frame.astype({column_name: np.int64})


def cast_to_short(data_frame: pd.DataFrame, column_name: str):
    """
    Casts given column of dataframe to short(np.int16).
    """
    return data_frame.astype({column_name: np.int16})


def cast_to_double(data_frame: pd.DataFrame, column_name: str):
    """
    Casts given column of dataframe to double(np.float64).
    """
    return data_frame.astype({column_name: np.float64})


def cast_to_float(data_frame: pd.DataFrame, column_name: str):
    """
    Casts given column of dataframe to float(np.float32).
    """
    return data_frame.astype({column_name: np.float32})


def cast_to_string(data_frame: pd.DataFrame, column_name: str):
    """
    Casts given column of dataframe to string.
    """
    data_frame[column_name] = [i if i is None else str(i) for i in data_frame[column_name]]
    return data_frame


def get_column_native_type(data_frame: pd.DataFrame, col: str):
    """Returns ColumnNativeType for given pandas DataFrame column"""
    dtype = data_frame.dtypes[col]
    return get_native_type_from_dtype(dtype)


def get_native_type_from_dtype(dtype):
    """Returns ColumnNativeType for given dtype"""
    if is_string_dtype(dtype):
        return ColumnNativeType.String
    if is_numeric_dtype(dtype):
        return ColumnNativeType.Numeric
    if is_datetime64_any_dtype(dtype):
        return ColumnNativeType.Date
    return ColumnNativeType.Other


def get_numerical_column_names(data_frame: pd.DataFrame) -> List[str]:
    """Returns numerical columns in frame"""
    return [col for col in data_frame.columns if get_column_native_type(data_frame, col) == ColumnNativeType.Numeric]


def get_correlation_matrix_memory_efficient(data_frame: pd.DataFrame, columns: List = None,
                                            n_threads: int = None, only_half_filled: bool = False,
                                            diagonal_values: float = 1.0) -> pd.DataFrame:
    """
    Generate correlation matrix for given frame

    :param List[str] columns: If given, only correlations of given columns will be calculated
    :param int n_threads: Number of threads to use on parallel correlation calculation.
        You can reduce this to reduce memory usage, but execution time will increase.
    :param bool only_half_filled: Only top triangle of the correlation matrix is filled
    :param float diagonal_values: Correlations for same column pairs will be filled with this value
    """
    columns = data_frame.columns.tolist() if columns is None else columns
    num_cols = len(columns)
    corr_matrix = np.empty((num_cols, num_cols))
    corr_matrix.fill(np.nan)
    col_pairs = [(i, j) for i in range(num_cols) for j in range(i, num_cols)]

    def calc_corr_and_fill_matrix(col1, col2):
        if col1 == col2:
            if only_half_filled:
                return
            correlation = diagonal_values
        else:
            correlation = np.corrcoef(data_frame[columns[col1]], data_frame[columns[col2]])[0][1]
        corr_matrix[col1][col2] = correlation
        if not only_half_filled:
            corr_matrix[col2][col1] = correlation

    if n_threads is not None:
        parallel_execution_helper.execute_parallel(col_pairs,
                                                   calc_corr_and_fill_matrix, num_jobs=n_threads,
                                                   require_shared_memory=True)
    else:
        parallel_execution_helper.execute_parallel(col_pairs,
                                                   calc_corr_and_fill_matrix, require_shared_memory=True)
    return pd.DataFrame(corr_matrix, columns=columns).set_index(pd.Index(columns))



"""Includes MulticlassReporterService class."""
from typing import List
import numpy as np
import pandas as pd
from pandas.api.types import is_numeric_dtype
from sklearn.metrics import accuracy_score, confusion_matrix, classification_report, roc_auc_score
from sklearn.preprocessing import label_binarize

from organon.ml.reporting.domain.objects.multiclass_report import MulticlassReport
from organon.ml.reporting.domain.services.base_reporter_service import BaseReporterService
from organon.ml.reporting.settings.objects.multiclass_reporter_settings import MultiClassReporterSettings


class MulticlassReporterService(BaseReporterService[MulticlassReport]):
    """Reporter service for multiclass classification"""

    def execute(self, settings: MultiClassReporterSettings) -> MulticlassReport:
        """returns multiclass report"""

        report = MulticlassReport()
        id_str_exists = settings.id_str_column is not None
        split_col_exists = settings.split_column is not None
        group_by_columns = []
        if id_str_exists:
            group_by_columns.append(settings.id_str_column)
        if split_col_exists:
            group_by_columns.append(settings.split_column)

        scores_table, confusion_matrix_dict, classification_report_dict = self._generate_reports(settings,
                                                                                                 id_str_exists,
                                                                                                 split_col_exists,
                                                                                                 group_by_columns)
        report.scores = scores_table
        report.confusion_matrices = confusion_matrix_dict
        report.classification_reports = classification_report_dict
        return report

    def _generate_reports(self, settings: MultiClassReporterSettings, id_str_exists: bool, split_col_exists: bool,
                          group_by_columns: List[str]):
        score_table_cols = self._initialize_acc_and_roc_table_cols(id_str_exists, split_col_exists,
                                                                               "Accuracy", "Roc-Auc Score")

        confusion_matrix_dict = {}
        classification_report_dict = {}
        roc_score = None
        scores_list = []
        score_column = settings.score_column if settings.probability_values is None else "Class Predicts"
        if settings.probability_values is not None:
            if settings.ordered_class_names is None:
                raise ValueError("When probability_values were given, ordered_class_names cannot be empty.")
            MulticlassReporterService._check_probability_values_data_length_match(settings)
            MulticlassReporterService._check_class_names_and_probability_values_match(settings)
            class_predicts = MulticlassReporterService._convert_probs_to_class_score(settings)
            settings.data["Class Predicts"] = class_predicts
        MulticlassReporterService._check_data_columns(settings, score_column)
        if group_by_columns:
            for groups, indices in settings.data.groupby(group_by_columns).indices.items():
                index_vals = list(set(list(settings.data.loc[indices][settings.target_column].unique()) +
                                      list(settings.data.loc[indices][score_column].unique())))
                index_vals.sort()
                group_tuple = groups if isinstance(groups, tuple) else (groups,)
                group_frame = settings.data.loc[indices]

                if settings.probability_values is not None:
                    roc_score = MulticlassReporterService._calculate_roc_auc_score(settings)

                confusion_matrix_dict[group_tuple] = pd.DataFrame(
                    data=confusion_matrix(group_frame[settings.target_column], group_frame[score_column],
                                          labels=index_vals), columns=index_vals, index=index_vals)
                classification_report_dict[group_tuple] = pd.DataFrame(
                    classification_report(group_frame[settings.target_column], group_frame[score_column],
                                          labels=index_vals, output_dict=True, zero_division=1)).transpose()

                acc_score = accuracy_score(group_frame[settings.target_column], group_frame[score_column])
                scores_list.append([*list(group_tuple), acc_score, roc_score])

            scores_df = pd.DataFrame(data=scores_list,
                                     columns=score_table_cols)

            if score_column == "Class Predicts":
                settings.data.drop(score_column, inplace=True, axis=1)

            return scores_df, confusion_matrix_dict, classification_report_dict

        index_vals = settings.data[settings.target_column].unique()

        if settings.probability_values is not None:
            roc_score = MulticlassReporterService._calculate_roc_auc_score(settings)

        confusion_matrix_dict[("ALL",)] = pd.DataFrame(
            data=confusion_matrix(settings.data[settings.target_column], settings.data[score_column],
                                  labels=index_vals), columns=index_vals, index=index_vals)
        classification_report_dict[("ALL",)] = pd.DataFrame(
            classification_report(settings.data[settings.target_column], settings.data[score_column],
                                  labels=index_vals, output_dict=True, zero_division=1)).transpose()

        scores_df = pd.DataFrame(data={"Accuracy": [accuracy_score(settings.data[settings.target_column],
                                                                   settings.data[score_column])],
                                       "Roc-Auc Score": [roc_score]})
        if score_column == "Class Predicts":
            settings.data.drop(score_column, inplace=True, axis=1)

        return scores_df, confusion_matrix_dict, classification_report_dict

    @staticmethod
    def _initialize_acc_and_roc_table_cols(id_str_exists: bool, split_col_exists: bool, acc_col: str, roc_col: str):
        columns = [acc_col, roc_col]
        if split_col_exists:
            columns.insert(0, "Data")
        if id_str_exists:
            columns.insert(0, "IdStr")
        return columns

    @staticmethod
    def _convert_probs_to_class_score(settings: MultiClassReporterSettings):
        """convert probability values to class predictions, and store proba values"""
        class_predictions = np.argmax(settings.probability_values, axis=1)
        arranged_class_predictions = []
        for class_name in class_predictions:
            arranged_class_name = settings.ordered_class_names.index(class_name)
            arranged_class_predictions.append(arranged_class_name)
        return pd.Series(arranged_class_predictions, index=settings.data.index)

    @staticmethod
    def _calculate_roc_auc_score(settings: MultiClassReporterSettings):
        """calculate roc auc score if predicted column has predict_proba values"""
        if len(np.unique(settings.data[settings.target_column])) > 1:
            y_test = label_binarize(settings.data[settings.target_column], classes=settings.ordered_class_names)
            scores = roc_auc_score(y_test, settings.probability_values, average='weighted',
                                   multi_class='ovr')
            return scores
        scores = None
        return scores

    @staticmethod
    def _check_data_columns(settings: MultiClassReporterSettings, score_column: str):
        """check if target and predicted columns includes string value if it does, raise error"""
        if not is_numeric_dtype(settings.data[settings.target_column]):
            raise ValueError("Target Column can only contain numeric values.")
        if not is_numeric_dtype(settings.data[score_column]):
            raise ValueError("Prediction Column can only contain numeric values.")

    @staticmethod
    def _check_class_names_and_probability_values_match(settings: MultiClassReporterSettings):
        """raise error if class order names array shape and prob values shapes mismatch"""
        if settings.probability_values is not None:
            num_classes = len(settings.ordered_class_names)
            if settings.probability_values.shape[1] != num_classes:
                raise ValueError("Given class names array shape mismatches with probability_values array.")

    @staticmethod
    def _check_probability_values_data_length_match(settings: MultiClassReporterSettings):
        """raise error if data target_column's length mismatches probability_values length"""
        if settings.probability_values is not None:
            if settings.probability_values.shape[0] != len(settings.data):
                raise ValueError("Given probability values mismatches with target column's length.")



"""todo"""
from typing import List

from organon.idq.domain.businessobjects.dq_calculation_result import DqCalculationResult
from organon.idq.domain.businessobjects.dq_comparison_result import DqComparisonResult
from organon.idq.domain.settings.dq_comparison_column_info import DqComparisonColumnInfo


class PsiComparer:
    """todo"""

    @staticmethod
    def nominal_columns_psi_comparer(controls: List[DqCalculationResult],
                                     t_calc_result: DqCalculationResult,
                                     test_nominal_cols: List[DqComparisonColumnInfo],
                                     bmh: int, psi_yellow: float, psi_red: float) -> List[DqComparisonResult]:
        """todo"""

    @staticmethod
    def numerical_columns_psi_comparer(controls: List[DqCalculationResult],
                                       t_calc_result: DqCalculationResult,
                                       test_nominal_cols: List[DqComparisonColumnInfo],
                                       bmh: int, psi_yellow: float, psi_red: float) -> List[DqComparisonResult]:
        """todo"""
        # frame = t_calc_result.sample_data  # c# taki frame argümanı yerine



"""Includes MailNotificationSettings class."""
from typing import List


class MailNotificationSettings:
    """"Settings for mail notification."""
    def __init__(self):
        self.addresses: List[str] = None
        self.cc_addresses: List[str] = None



"""Includes ModelParameterSettings Class"""
from typing import List


class ModelParameterSettings:
    """Class for model parameters"""

    def __init__(self, batch_size: int = 10, epochs: int = 20, metrics: List[str] = None):
        self.batch_size: int = batch_size
        self.epochs: int = epochs
        self.metrics: List[str] = metrics



"""Includes SingleSourceAlertInfo class"""
from organon.idq.domain.reporting.objects.base_alert_info import BaseAlertInfo


class SingleSourceAlertInfo(BaseAlertInfo):
    """Alert Info for SingleSource """



"""Includes GBMRegressor class."""
from sklearn.ensemble import GradientBoostingRegressor

from organon.ml.modelling.algorithms.core.abstractions.base_sklearn_regressor import BaseSklearnRegressor


class GBMRegressor(BaseSklearnRegressor[GradientBoostingRegressor]):
    """Wrapper for sklearn GradientBoostingRegressor"""

    @property
    def regressor_class(self):
        return GradientBoostingRegressor



"""todo"""
from typing import Optional, Dict

from organon.idq.domain.businessobjects.statistics.dq_categorical_statistics import DqCategoricalStatistics
from organon.idq.domain.businessobjects.statistics.dq_population_numerical_statistics import \
    DqPopulationNumericalStatistics


class PopulationStatistics:
    """todo"""
    def __init__(self):
        self.nominal_statistics: Optional[Dict[str, DqCategoricalStatistics]] = None
        self.numerical_statistics: Optional[Dict[str, DqPopulationNumericalStatistics]] = None



"""Includes ColumnMetadata class."""
from typing import List


class DqColumnMetadata:
    """Column metadata settings"""

    def __init__(self):
        self.column_name: str = None
        self.default_values: List[str] = None
        self.inclusion_flag: bool = None



"""
This module includes TransactionFileDescriptor class.
"""

from organon.afe.domain.common.reader_helper import get_values_from_kwargs
from organon.afe.domain.settings.base_trx_descriptor import BaseTrxDescriptor
from organon.afe.domain.settings.feature_generation_settings import FeatureGenerationSettings


class TrxDescriptor(BaseTrxDescriptor):
    """
    Class for information about transaction file or database table
    """
    ATTR_DICT = {
        "feature_gen_setting": FeatureGenerationSettings
    }
    ATTR_DICT.update(BaseTrxDescriptor.ATTR_DICT)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.feature_gen_setting: FeatureGenerationSettings = None

        get_values_from_kwargs(self, self.ATTR_DICT, kwargs)

    def get_all_dimension_and_quantity_columns(self):
        """Returns sets of all dimension and quantity columns defined in date column settings"""
        all_d_cols = set()
        all_q_cols = set()
        all_d_cols.update(self.feature_gen_setting.dimension_columns)
        all_q_cols.update(self.feature_gen_setting.quantity_columns)
        return all_d_cols, all_q_cols



"""Includes LassoSelection class."""
from typing import Union, Optional

import pandas as pd

from organon.ml.common.enums.target_type import TargetType
from organon.ml.common.helpers.user_input_service_helper import get_enum
from organon.ml.feature_selection.domain.objects.lasso_selection_output import LassoSelectionOutput
from organon.ml.feature_selection.domain.objects.settings.lasso_selection_settings import LassoSelectionSettings
from organon.ml.feature_selection.domain.services.lasso_selection_service import LassoSelectionService
from organon.ml.feature_selection.services.abstractions.base_supervised_feature_selecter import \
    BaseSupervisedFeatureSelecter


class LassoSelection(BaseSupervisedFeatureSelecter):
    """Feature selecter which utilizes Lasso for selection"""

    def __init__(self, data: pd.DataFrame, target: Union[pd.DataFrame, pd.Series], target_type:str, **kwargs):
        super().__init__(data, target)
        target = self._get_target_as_df(target)
        self.settings = LassoSelectionSettings(data, target, target_type=get_enum(target_type, TargetType),
                                               lasso_args=kwargs)
        self.output: Optional[LassoSelectionOutput] = None

    def _run(self):
        service = LassoSelectionService()
        self.output = service.run_selection(self.settings)
        return self.output



"""
This module includes AfeProcessSettings class.
"""
import numpy as np

from organon.afe.domain.common.reader_helper import get_values_from_kwargs


class AfeProcessSettings:
    """
    Settings for Automated Feature Extraction process
    """
    ATTR_DICT = {
        "process_id": str,
        "number_of_cores": int,
        "max_memory_size": np.float64,
        "max_execution_time_in_seconds": np.int64
    }

    def __init__(self, **kwargs):
        self.process_id: str = None
        self.number_of_cores: int = None
        self.max_memory_size: np.float64 = None
        self.max_execution_time_in_seconds: np.int64 = None

        get_values_from_kwargs(self, self.ATTR_DICT, kwargs)



"""This module includes StableColumnControl class."""
import math
from typing import List, Dict

from organon.fl.core.enums.column_native_type import ColumnNativeType
from organon.fl.core.helpers.string_helper import is_null_or_empty
from organon.fl.logging.helpers.log_helper import LogHelper
from organon.idq.core.enums.data_entity_type import DataEntityType
from organon.idq.core.enums.dq_comparison_result_code import DqComparisonResultCode
from organon.idq.domain.businessobjects.dq_comparison_result import DqComparisonResult
from organon.idq.domain.businessobjects.statistics.dq_categorical_statistics import DqCategoricalStatistics
from organon.idq.domain.businessobjects.statistics.dq_sample_numerical_statistics import DqSampleNumericalStatistics
from organon.idq.domain.controls.base_dq_control import BaseDqControl
from organon.idq.domain.enums.dq_control_type import DqControlType
from organon.idq.domain.enums.dq_test_group_type import DqTestGroupType


class StableColumnControl(BaseDqControl):
    """Class for Stable Column Control."""

    @staticmethod
    def get_control_type() -> DqControlType:
        """:returns control type"""
        return DqControlType.STABLE_COLUMN

    @classmethod
    def get_test_group_type(cls) -> DqTestGroupType:
        """:returns group type of dq control"""
        return DqTestGroupType.RULE_BASED_CONTROLS_COLUMN_SET

    def get_description(self) -> str:
        """:returns description of control"""
        return "Stable Column Control"

    def _execute_control(self) -> List[DqComparisonResult]:
        """:return: list of comparison results"""
        self.__validate_parameters(self.comp_context)

        nominal_comp_cols = self.comp_context.get_nominal_comparison_columns()
        nominal_col_names = [comp_col.column_name for comp_col in nominal_comp_cols]
        sample_nominal_stats = {col: stat for col, stat in
                                self.comp_context.test_calc_result.sample_stats.nominal_statistics.items()
                                if col in nominal_col_names}
        nominal_cardinalities = {stat: sample_nominal_stats[stat].cardinality for stat in sample_nominal_stats}
        nominal_freqs = {stat: sample_nominal_stats[stat].frequencies for stat in sample_nominal_stats}

        sample_numeric_stats = self.comp_context.test_calc_result.sample_stats.numerical_statistics
        numeric_cardinalities = {stat: sample_numeric_stats[stat].cardinality for stat in sample_numeric_stats}
        numeric_freqs = {stat: sample_numeric_stats[stat].frequencies for stat in sample_numeric_stats}

        all_columns_cardinality = numeric_cardinalities
        all_columns_cardinality.update(nominal_cardinalities)
        all_columns_freq = numeric_freqs
        all_columns_freq.update(nominal_freqs)

        results = []
        results.extend(self.__control_cols_for_stability())

        max_cardinality = self.comp_context.comparison_parameters.maximum_nom_cardinality
        min_cardinality = self.comp_context.comparison_parameters.minimum_cardinality
        if max_cardinality is None:
            LogHelper.info("Max cardinality is None, skipping max cardinality control.")
        else:
            if any(column_distinct_count >= max_cardinality for column_distinct_count in
                   nominal_cardinalities.values()):
                results.extend(self.__column_max_cardinality_control(nominal_cardinalities, max_cardinality))

        if min_cardinality is None:
            LogHelper.info("Min cardinality is None, skipping min cardinality control.")
        else:
            if any(column_distinct_count <= min_cardinality for column_distinct_count in
                   all_columns_cardinality.values()):
                results.extend(self.__column_min_cardinality_control(all_columns_cardinality, min_cardinality))
        return results

    @staticmethod
    def __validate_parameters(comp_context):
        if comp_context.test_calc_result is None:
            raise ValueError("T-Calculation result is None.")
        if comp_context.test_calc_result.sample_stats is None:
            raise ValueError("T-Calculation result sample stats is None.")
        if comp_context.test_calc_result.sample_stats.numerical_statistics is None:
            raise ValueError("T-Calculation result numeric sample stats is None.")
        if comp_context.test_calc_result.sample_stats.nominal_statistics is None:
            raise ValueError("T-Calculation result nominal sample stats is None.")
        if comp_context.comparison_parameters is None:
            raise ValueError("T-Calculation comparison parameters is None.")

    def __control_cols_for_stability(self):
        results = []
        sample_numeric_stats = self.comp_context.test_calc_result.sample_stats.numerical_statistics
        for comp_col in self.comp_context.get_comparison_cols_by_native_type(ColumnNativeType.Numeric):
            col = comp_col.column_name
            if comp_col.column_name not in sample_numeric_stats:
                continue
            stats = sample_numeric_stats[col]
            if stats.missing_values is None:
                continue
            self.__control_all_zero(results, col, stats)
            self.__control_numeric_all_null(results, col, stats)
            self.__control_numeric_constant(results, col, stats)

        nominal_stats = self.comp_context.test_calc_result.sample_stats.nominal_statistics
        for comp_col in self.comp_context.get_nominal_comparison_columns():
            if comp_col.column_name in nominal_stats:
                self.__control_nominal_all_null_or_constant(results, comp_col.column_name,
                                                            nominal_stats[comp_col.column_name])
        return results

    def __control_all_zero(self, results: List[DqComparisonResult], col: str, stats: DqSampleNumericalStatistics):
        default_values = stats.missing_values
        default_frequencies = stats.missing_values_frequencies
        if 0 in default_values:
            n_val = stats.n_val
            n_miss = stats.n_miss
            zero_count = default_frequencies[0] if 0 in default_frequencies else 0
            if n_val > 0 and n_val == n_miss and n_val == zero_count:
                comp_result = self._get_comparison_result(
                    data_entity=DataEntityType.COLUMN,
                    data_entity_name=col,
                    result_code=DqComparisonResultCode.COLUMN_VALUES_ARE_ALL_ZERO,
                    message=f"Column values consist of only zeros for the column: {col}",
                    property_code="ColumnValuesAreAllZero",
                )
                comp_result.property_key_nominal = "true"
                results.append(comp_result)
        else:
            min_val = stats.min
            max_val = stats.max
            if min_val == 0 and max_val == 0:
                comp_result = self._get_comparison_result(
                    data_entity=DataEntityType.COLUMN,
                    data_entity_name=col,
                    result_code=DqComparisonResultCode.COLUMN_VALUES_ARE_ALL_ZERO,
                    message=f"Non-default column values consist of only zeros for the column: {col}",
                    property_code="ColumnValuesAreAllZero"
                )
                comp_result.property_key_nominal = "true"
                results.append(comp_result)

    def __control_numeric_all_null(self, results: List[DqComparisonResult], col: str,
                                   stats: DqSampleNumericalStatistics):
        default_values = stats.missing_values
        default_frequencies = stats.missing_values_frequencies
        if None in default_values:
            n_val = stats.n_val
            n_miss = stats.n_miss
            nan_count = default_frequencies[None] if None in default_frequencies else 0
            if n_val > 0 and n_val == n_miss and n_val == nan_count:
                comp_result = self._get_comparison_result(
                    data_entity=DataEntityType.COLUMN,
                    data_entity_name=col,
                    result_code=DqComparisonResultCode.COLUMN_VALUES_ARE_ALL_NULL,
                    message=f"Column values consist of only NULL values for the column: {col}",
                    property_code="ColumnValuesAreAllNull",
                )
                comp_result.property_key_nominal = "true"
                results.append(comp_result)
        else:
            min_val = stats.min
            max_val = stats.max
            if math.isnan(min_val) and math.isnan(max_val):
                comp_result = self._get_comparison_result(
                    data_entity=DataEntityType.COLUMN,
                    data_entity_name=col,
                    result_code=DqComparisonResultCode.COLUMN_VALUES_ARE_ALL_NULL,
                    message=f"Non-default column values consist of only NULL values for the column: {col}",
                    property_code="ColumnValuesAreAllNull"
                )
                comp_result.property_key_nominal = "true"
                results.append(comp_result)

    def __control_numeric_constant(self, results: List[DqComparisonResult], col: str,
                                   stats: DqSampleNumericalStatistics):
        if stats.cardinality == 1 and stats.frequencies is not None and len(stats.frequencies) == 1:
            single_val = next(iter(stats.frequencies.keys()))
            comp_result = self._get_comparison_result(
                data_entity=DataEntityType.COLUMN,
                data_entity_name=col,
                result_code=DqComparisonResultCode.COLUMN_IS_CONSTANT,
                message=f"Column consists of a single value ({single_val}) for the column: {col}",
                property_code="ColumnIsConstant"
            )
            comp_result.property_key_nominal = "true"
            results.append(comp_result)

    def __control_nominal_all_null_or_constant(self, results: List[DqComparisonResult], col: str,
                                               stats: DqCategoricalStatistics):
        frequencies = stats.frequencies
        cardinality = len(frequencies)
        if cardinality == 1:
            single_val = next(iter(frequencies.keys()))
            val = str(single_val) if single_val is not None else None
            if is_null_or_empty(val):
                comp_result = self._get_comparison_result(
                    data_entity=DataEntityType.COLUMN,
                    data_entity_name=col,
                    result_code=DqComparisonResultCode.COLUMN_VALUES_ARE_ALL_NULL,
                    message=f"Column consists of only NULL values for the column: {col}",
                    property_code="ColumnValuesAreAllNull"
                )
                comp_result.property_key_nominal = "true"
                results.append(comp_result)
            else:
                comp_result = self._get_comparison_result(
                    data_entity=DataEntityType.COLUMN,
                    data_entity_name=col,
                    result_code=DqComparisonResultCode.COLUMN_IS_CONSTANT,
                    message=f"Column consists of a single value ({val}) for the column: {col}",
                    property_code="ColumnIsConstant"
                )
                comp_result.property_key_nominal = "true"
                results.append(comp_result)

    @classmethod
    def __column_max_cardinality_control(cls, unique_val_nums_list: Dict[str, int], max_cardinality: int) -> \
            List[DqComparisonResult]:
        results = []
        above_max_cols_names = [col for col, unq_cnt in unique_val_nums_list.items() if unq_cnt >= max_cardinality]
        for col in above_max_cols_names:
            comp_result = cls._get_comparison_result(
                data_entity=DataEntityType.COLUMN,
                data_entity_name=col,
                result_code=DqComparisonResultCode.CARDINALITY_EXCEEDS_MAXIMUM_CARDINALITY,
                property_code="MaximumCardinalityForTheColumn",
                message=f"Column cardinality is greater than or equal to {max_cardinality} for the column: {col}"
            )
            comp_result.property_key_numeric = max_cardinality
            results.append(comp_result)
        return results

    @classmethod
    def __column_min_cardinality_control(cls, unique_val_nums_list: Dict[str, int], min_cardinality: int) -> \
            List[DqComparisonResult]:
        results = []
        below_min_cols_names = [col for col, unq_cnt in unique_val_nums_list.items() if unq_cnt <= min_cardinality]
        for col in below_min_cols_names:
            comp_result = cls._get_comparison_result(
                data_entity=DataEntityType.COLUMN,
                data_entity_name=col,
                result_code=DqComparisonResultCode.CARDINALITY_IS_BELOW_MINIMUM_THRESHOLD,
                property_code="MinimumCardinalityForTheColumn",
                message=f"Column cardinality is less than or equal to {min_cardinality} for the column: {col}"
            )
            comp_result.property_key_numeric = min_cardinality
            results.append(comp_result)
        return results



"""Includes OptimizerType enum class"""
from enum import Enum, auto


class OptimizerType(Enum):
    """Optimizer types"""
    ADAM = auto()
    ADADELTA = auto()
    ADAGRAD = auto()
    NADAM = auto()
    FTRL = auto()
    ADAMAX = auto()
    SGD = auto()
    RMSPROP = auto()
    


"""
This module includes ModelSettings class.
"""

from organon.afe.domain.common.reader_helper import get_values_from_kwargs


class ModelSettings:
    """
    Model settings for supervised feature selection
    """
    ATTR_DICT = {
        "model_params": dict,
        "model_fit_params": dict,
        "reduction_coverage": float
    }

    def __init__(self, **kwargs):
        self.model_params: dict = None
        self.model_fit_params: dict = None
        self.reduction_coverage: float = None

        get_values_from_kwargs(self, self.ATTR_DICT, kwargs)



"""Includes FeatureReduction class."""
from typing import List

import pandas as pd

from organon.ml.feature_reduction.domain.objects.feature_reduction_output import FeatureReductionOutput

from organon.ml.feature_reduction.domain.services.feature_reduction_service import FeatureReductionService
from organon.ml.feature_reduction.settings.feature_reduction_user_input_service import FeatureReductionUserInputService
from organon.ml.feature_reduction.settings.objects.user_feature_reduction_settings import UserFeatureReductionSettings

class FeatureReduction:
    """Feature reduction user interface class"""

    def __init__(self):
        self._output: List[FeatureReductionOutput] = None
        self._input_service = FeatureReductionUserInputService()
        self.settings: UserFeatureReductionSettings = UserFeatureReductionSettings()

    def execute(self, data: pd.DataFrame, drop_cols: bool, included_reduction_types: List[str] = None,
                target_type: str = None, target_column_name: str = None, performance_metric: str = None,
                excluded_columns: List[str] = None
                ) -> \
            List[FeatureReductionOutput]:
        """
        Execute feature reduction with given parameters
        :param included_reduction_types:
        :param data:
        :param drop_cols:
        :param target_type:
        :param target_column_name:
        :param performance_metric:
        :return:
        """
        # pylint: disable=too-many-arguments
        self.settings.data = data
        self.settings.included_reduction_types = included_reduction_types
        self.settings.target_type = target_type
        self.settings.target_column_name = target_column_name
        self.settings.performance_metric = performance_metric
        self.settings.excluded_columns = excluded_columns
        feature_reduction_settings = self._input_service.generate_feature_reduction_settings(self.settings)
        self._output = FeatureReductionService.execute(feature_reduction_settings, self.settings.data, drop_cols)

        return self._output

    def set_null_feature_reduction_settings(self, null_ratio_threshold: float = 0.99):
        """
        Set null_ratio_threshold parameter for null feature reduction entered by user
        :param null_ratio_threshold:
        :return:
        """
        self.settings.null_ratio_threshold = null_ratio_threshold

    def set_high_correlated_feature_reduction_settings(self, correlation_threshold: float = 0.99,
                                                       random_state=None):
        """
        Set correlation_threshold parameter for high correlated feature reduction
        :param correlation_threshold:
        :param random_state:
        :return:
        """
        self.settings.correlation_threshold = correlation_threshold
        self.settings.random_state = random_state

    def set_univariate_performance_feature_reduction_settings(self, univariate_performance_threshold: float = None,
                                                              random_state=None):
        """
        Set univariate_performance_threshold parameter for univariate performance feature reduction
        :param univariate_performance_threshold:
        :param random_state:
        :return:
        """
        self.settings.univariate_performance_threshold = univariate_performance_threshold
        self.settings.random_state = random_state

    def set_similar_distribution_feature_reduction_settings(self, nunique_count: int = 20, random_state=None):
        """
        Set nunique_count parameter for similar distribution feature reduction

        :param nunique_count:
        :param random_state:
        :return:
        """
        self.settings.nunique_count = nunique_count
        self.settings.random_state = random_state



"""todo"""
from typing import List

from organon.idq.domain.businessobjects.dq_calculation_result import DqCalculationResult
from organon.idq.domain.businessobjects.dq_comparison_result import DqComparisonResult
from organon.idq.domain.settings.dq_comparison_column_info import DqComparisonColumnInfo


class ColumnValuesComparer:
    """todo"""

    @staticmethod
    def compare_unexpected_nominal_values(controls: List[DqCalculationResult],
                                          t_calc_result: DqCalculationResult,
                                          test_nominal_cols: List[DqComparisonColumnInfo], bmh: int) \
            -> List[DqComparisonResult]:
        """todo"""

    @staticmethod
    def compare_unexpected_numerical_values(controls: List[DqCalculationResult],
                                            t_calc_result: DqCalculationResult,
                                            test_numerical_cols: List[DqComparisonColumnInfo], bmh: int) \
            -> List[DqComparisonResult]:
        """todo"""

    @staticmethod
    def test_nominal_column_values(t_calc_result: DqCalculationResult,
                                   nominal_columns_collection: List[DqComparisonColumnInfo],
                                   minimum_cardinality: int) -> List[DqComparisonResult]:
        """todo"""

    @staticmethod
    def test_numerical_column_values(t_calc_result: DqCalculationResult,
                                     numerical_columns_collection: List[DqComparisonColumnInfo],
                                     minimum_cardinality: int) -> List[DqComparisonResult]:
        """todo"""



"""
Built-in datasets for demonstration, educational and test purposes.
"""
import os

import pandas


def base_df():
    """
    Returns: "base" pandas.DataFrame
    """
    data_df = _get_dataset("base_df")
    return data_df.set_index(data_df.columns[0])  # pylint: disable=no-member


def score_df():
    """
    Returns: "score" pandas.DataFrame
    """
    data_df = _get_dataset("score_df")
    return data_df.set_index(data_df.columns[0])  # pylint: disable=no-member


def cash_df():
    """
    Returns: "cash" pandas.DataFrame
    """
    data_df = _get_dataset("cash_df")
    return data_df.set_index(data_df.columns[0])  # pylint: disable=no-member


def _get_dataset(dataset_name):
    return pandas.read_csv(
        os.path.join(
            os.path.dirname(os.path.dirname(__file__)),
            "data",
            dataset_name + ".zip",
        )
    )



"""Includes LabeledDirDataSourceHandler class"""
import os
from typing import List

import numpy as np
from pandas import DataFrame

from organon.ml.common.enums.classification_type import ClassificationType
from organon.ml.common.enums.color_type import ColorType
from organon.ml.object_classification.common.helpers import \
    throw_val_ex_with_log
from organon.ml.object_classification.common.object_classification_constants import \
    ObjectClassificationConstants
from organon.ml.object_classification.domain.data_sources.base_data_source_handler import \
    BaseDataSourceHandler
from organon.ml.object_classification.domain.objects.data_source_settings import \
    DataSourceSettings


class LabeledDirDataSourceHandler(BaseDataSourceHandler[str]):
    """Class for handling labeled directory input"""

    def __init__(self, data: str, settings: DataSourceSettings):
        super().__init__(data, settings)
        self.filenames: List[str] = None

    def process_data(self):
        import tensorflow as tf # pylint: disable=import-outside-toplevel
        self.filenames = self._get_labeled_dir_filenames(self.data)

        label_mode = "binary" if self.settings.clf_type == ClassificationType.BINARY else "categorical"
        self._dataset = tf.keras.utils.image_dataset_from_directory(directory=self.data,
                                                                    batch_size=self.settings.batch_size,
                                                                    image_size=self.settings.image_size,
                                                                    shuffle=False, seed=self.settings.random_seed,
                                                                    color_mode=self._get_color_parameter(),
                                                                    label_mode=label_mode)

    def initialize_probs_df(self) -> DataFrame:
        return DataFrame({ObjectClassificationConstants.FILENAME_COL_NAME: self.filenames,
                          ObjectClassificationConstants.LABELS_COL_NAME: self.get_labels()})

    def _get_color_parameter(self) -> str:
        if self.settings.color_mode == ColorType.RGB:
            color_mode = "rgb"
        elif self.settings.color_mode == ColorType.RGBA:
            color_mode = "rgba"
        elif self.settings.color_mode == ColorType.GRAY:
            color_mode = "grayscale"
        else:
            return throw_val_ex_with_log("Color mode is not defined!")
        return color_mode

    @staticmethod
    def _get_labeled_dir_filenames(directory: str) -> List[str]:
        files = [file for _, _, file in os.walk(
            directory)]  # image_dataset_from_directory ile aynı sırada geliyor, o fonksiyon da os.walk kullanıyor
        return list(np.concatenate(files).flat)

    def _get_binary_labels(self):
        labels = self._extract_labels()
        return labels.flatten()

    def _get_multiclass_labels(self):
        return np.argmax(np.concatenate(self._extract_labels(), axis=0), axis=1)



"""Includes SingleSourceReportHelperParams class."""

from organon.idq.domain.reporting.objects.base_dq_report_helper_params import BaseDqReportHelperParams


class SingleSourceReportHelperParams(BaseDqReportHelperParams):
    """Params for DQ report helper."""



"""Includes FlSortedList class."""
from typing import List, TypeVar, Generic

from sortedcontainers import SortedList as _SortedList

T = TypeVar('T')


class SortedList(Generic[T]):
    """A sorted list implementation using sortedcontainers.SortedList"""

    def __init__(self, values: List[T] = None):
        values = [] if values is None else values
        self._list = _SortedList(values)

    def add(self, value: T):
        """Add element to list"""
        self._list.add(value)

    def __getitem__(self, index):
        return self._list[index]

    def __len__(self):
        return len(self._list)

    def __iter__(self):
        return iter(self._list)

    def __str__(self):
        return str(self._list)

    def __repr__(self):
        return repr(self._list)



"""
This module includes BaseAfeSettingsReader class.
"""
from datetime import datetime
from typing import Union, List, Dict, TypeVar, Generic

import numpy as np
import pandas as pd

from organon.afe.domain.common.reader_helper import set_defaults_from_dict, setattr_if_none
from organon.afe.domain.enums.afe_date_column_type import AfeDateColumnType
from organon.afe.domain.enums.afe_learning_type import AfeLearningType
from organon.afe.domain.enums.afe_operator import AfeOperator
from organon.afe.domain.enums.afe_target_column_type import AfeTargetColumnType
from organon.afe.domain.enums.date_resolution import DateResolution
from organon.afe.domain.modelling.supervised.afe_supervised_algorithm_settings import AfeSupervisedAlgorithmSettings
from organon.afe.domain.modelling.unsupervised.afe_unsupervised_algorithm_settings import \
    AfeUnsupervisedAlgorithmSettings
from organon.afe.domain.settings.afe_algorithm_settings import AfeAlgorithmSettings
from organon.afe.domain.settings.afe_data_settings import AfeDataSettings
from organon.afe.domain.settings.afe_date_column import AfeDateColumn
from organon.afe.domain.settings.afe_modelling_settings import AfeModellingSettings
from organon.afe.domain.settings.afe_output_settings import AfeOutputSettings
from organon.afe.domain.settings.afe_process_settings import AfeProcessSettings
from organon.afe.domain.settings.afe_reading_settings import AfeDataReadingSettings
from organon.afe.domain.settings.afe_target_column import AfeTargetColumn
from organon.afe.domain.settings.auto_column_decider_settings import AutoColumnDeciderSettings
from organon.afe.domain.settings.base_afe_modelling_settings import BaseAfeModellingSettings
from organon.afe.domain.settings.base_afe_scoring_settings import BaseAfeScoringSettings
from organon.afe.domain.settings.base_afe_settings import BaseAfeSettings
from organon.afe.domain.settings.binary_target import BinaryTarget
from organon.afe.domain.settings.db_object_input import DbObjectInput
from organon.afe.domain.settings.feature_generation_settings import FeatureGenerationSettings
from organon.afe.domain.settings.model_settings import ModelSettings
from organon.afe.domain.settings.record_source import RecordSource
from organon.afe.domain.settings.target_descriptor import TargetDescriptor
from organon.afe.domain.settings.temporal_grid import TemporalGrid
from organon.afe.domain.settings.trx_descriptor import TrxDescriptor
from organon.fl.core.helpers import environment_info_helper

T = TypeVar("T", bound=BaseAfeModellingSettings)


class BaseAfeSettingsReader(Generic[T]):
    """
    Class for static methods on reading AfeModellingSettings from a file
    """

    def assign_default_values(self, modelling_settings: T):
        """
        Assigns default values to given AfeModellingSettings instance
        :param modelling_settings: AfeModellingSettings instance
        :type modelling_settings: BaseAfeModellingSettings
        """
        default_values_dict = self._get_default_values_dict()
        if modelling_settings.data_source_settings.auto_column_decider_settings is None:
            modelling_settings.data_source_settings.auto_column_decider_settings = AutoColumnDeciderSettings()

        set_defaults_from_dict(modelling_settings, default_values_dict)

        self.__set_binary_target_defaults(modelling_settings)

        self._set_algorithm_settings_defaults(modelling_settings)

        self._set_trx_descriptor_defaults(modelling_settings)
        self.__set_target_descriptor_defaults(modelling_settings)

        self._assign_base_afe_settings_defaults(modelling_settings)

    @classmethod
    def _get_default_values_dict(cls):
        return {
            "data_source_settings": {
                "entity_table_existence": False,
                "max_number_of_target_samples": 100000,
                "max_number_of_transaction_samples": np.iinfo(np.int64).max,
                "number_of_retry_for_empty_transaction": 10,
                "retry_interval_seconds_for_empty_transaction": 600,
                "auto_column_decider_settings": {
                    "sampling_ratio": 0.01,
                    "numeric_to_dimension": 5,
                    "dimension_distinct_cut_off": 0.95,
                    "use_dimension_columns": False,
                    "rejected_dimension_columns": [],
                    "use_quantity_columns": False,
                    "rejected_quantity_columns": []
                }
            },
            "output_settings": {
                "output_folder": "AfeOutput",
                "feature_name_prefix": "FTR",
                "enable_feature_lookup_output_to_csv": True,
                "enable_write_output": False,
                "enable_feature_space_output": False,
                "return_all_afe_columns": False,
                "enable_all_feature_lookup_output_to_csv": False
            }
        }

    @staticmethod
    def __set_binary_target_defaults(modelling_settings: T):
        try:
            target_column = modelling_settings.data_source_settings.target_descriptor.target_column
            if target_column.binary_target_info is None:
                target_column.binary_target_info = BinaryTarget()
            binary_target_info = target_column.binary_target_info
            if binary_target_info.positive_category is None and binary_target_info.negative_category is None \
                    and binary_target_info.indeterminate_category is None \
                    and binary_target_info.exclusion_category is None:
                binary_target_info.positive_category = "P"
                binary_target_info.negative_category = "N"
                binary_target_info.indeterminate_category = "I"
                binary_target_info.exclusion_category = "X"
        except AttributeError:
            pass

    def _set_algorithm_settings_defaults(self, modelling_settings: T):
        is_supervised = modelling_settings.afe_learning_type == AfeLearningType.Supervised
        if modelling_settings.algorithm_settings is None:
            if is_supervised:
                modelling_settings.algorithm_settings = AfeSupervisedAlgorithmSettings()
            else:
                modelling_settings.algorithm_settings = AfeUnsupervisedAlgorithmSettings()

        defaults = {
            "dimension_compression_ratio": 1.0,
            "dimension_max_cardinality": 200
        }
        set_defaults_from_dict(modelling_settings.algorithm_settings, defaults)

        if is_supervised:
            self.__set_supervised_algorithm_settings_defaults(modelling_settings)
        else:
            self.__set_unsupervised_algorithm_settings_defaults(modelling_settings)

    @staticmethod
    def __set_supervised_algorithm_settings_defaults(modelling_settings: T):

        defaults = {
            "training_percentage": 0.8,
            "min_data_in_leaf_and_sample_size_control_ratio": 0.1
        }
        model_defaults = {
            "model_params": {
                "learning_rate": 0.01,
                "n_estimators": 200,
                "min_data_in_leaf": 1
            },
            "model_fit_params": {
                "early_stopping_rounds": 5
            },
            "reduction_coverage": 0.5

        }
        final_model_defaults = {
            "model_params": {
                "learning_rate": 0.01,
                "n_estimators": 200,
                "min_data_in_leaf": 1
            },
            "model_fit_params": {
                "early_stopping_rounds": 5
            },
            "reduction_coverage": 0.5

        }

        algorithm_settings: AfeSupervisedAlgorithmSettings = modelling_settings.algorithm_settings
        set_defaults_from_dict(algorithm_settings, defaults)

        if algorithm_settings.model_settings is None:
            algorithm_settings.model_settings = ModelSettings()
        set_defaults_from_dict(algorithm_settings.model_settings, model_defaults)

        if algorithm_settings.final_model_settings is None:
            algorithm_settings.final_model_settings = ModelSettings()
        set_defaults_from_dict(algorithm_settings.final_model_settings, final_model_defaults)

    @staticmethod
    def __set_unsupervised_algorithm_settings_defaults(modelling_settings: T):
        defaults = {
            "bin_count": 4,
            "r_factor": 0.9,
            "max_column_count": 500,
            "is_logging": False
        }
        set_defaults_from_dict(modelling_settings.algorithm_settings, defaults)

    def _set_trx_descriptor_defaults(self, modelling_settings: T):
        # pylint: disable=no-self-use
        try:
            trx_descriptor = modelling_settings.data_source_settings.trx_descriptor
            if trx_descriptor.reading_settings is None:
                trx_descriptor.reading_settings = AfeDataReadingSettings()

            setattr_if_none(trx_descriptor.reading_settings, "number_of_rows_per_step", 100000)

        except AttributeError:
            pass

    @staticmethod
    def __set_target_descriptor_defaults(modelling_settings: T):
        try:
            target_descriptor = modelling_settings.data_source_settings.target_descriptor
            if target_descriptor.reading_settings is None:
                target_descriptor.reading_settings = AfeDataReadingSettings()

            setattr_if_none(target_descriptor.reading_settings, "number_of_rows_per_step", 100000)

        except AttributeError:
            pass

    def assign_default_values_for_scoring(self, scoring_settings: BaseAfeScoringSettings):
        """Assigns default values to scoring settings."""
        if scoring_settings.trx_reading_settings is None:
            scoring_settings.trx_reading_settings = AfeDataReadingSettings(number_of_rows_per_step=100000)

        self._assign_base_afe_settings_defaults(scoring_settings)

    @staticmethod
    def _assign_base_afe_settings_defaults(settings: BaseAfeSettings):
        """
        Assigns default values to given BaseAfeSettings instance
        :param settings: BaseAfeSettings instance
        :type settings: BaseAfeSettings
        """
        num_cores = environment_info_helper.get_total_cores()
        defaults_dict = {
            "process_settings": {
                "number_of_cores": num_cores - 1 if num_cores != 1 else 1
            }
        }
        if settings.process_settings is None:
            settings.process_settings = AfeProcessSettings()
        set_defaults_from_dict(settings, defaults_dict)

    @staticmethod
    def get_date_column(date_column_name: str = None,
                        date_column_type: Union[str, AfeDateColumnType] = None,
                        date_column_format: str = None,
                        date_column_db_format: str = None):
        """Generates AfeDateColumn instance"""
        if date_column_name is None:
            return None
        return AfeDateColumn(column_name=date_column_name, date_column_type=date_column_type,
                             custom_format=date_column_format, db_custom_format=date_column_db_format)

    def get_feature_generation_setting(self,
                                       temporal_grids: List[Union[TemporalGrid, Dict]] = None,
                                       date_resolution: Union[DateResolution, str] = None,
                                       date_column_name: str = None,
                                       date_column_type: Union[str, AfeDateColumnType] = None,
                                       date_column_format: str = None,
                                       dimension_columns: List[str] = None,
                                       quantity_columns: List[str] = None,
                                       included_operators: List[Union[AfeOperator, str]] = None,
                                       date_offset: int = None,
                                       max_observation_date: datetime = None
                                       ) -> FeatureGenerationSettings:
        """
        Generates FeatureGenerationSettings instance.
        """

        # pylint: disable=too-many-arguments

        date_column = self.get_date_column(date_column_name=date_column_name,
                                           date_column_type=date_column_type,
                                           date_column_format=date_column_format)

        return FeatureGenerationSettings(temporal_grids=temporal_grids,
                                         date_resolution=date_resolution,
                                         date_column=date_column,
                                         dimension_columns=dimension_columns,
                                         quantity_columns=quantity_columns,
                                         included_operators=included_operators,
                                         date_offset=date_offset,
                                         max_observation_date=max_observation_date
                                         )

    @staticmethod
    def get_auto_column_decider_settings(sampling_ratio: float,
                                         numeric_to_dimension: int,
                                         dimension_distinct_cut_off: float,
                                         use_dimension_columns: bool,
                                         rejected_dimension_columns: List[str],
                                         use_quantity_columns: bool,
                                         rejected_quantity_columns: List[str]):
        """Generates AutoColumnDeciderSettings instance"""
        return AutoColumnDeciderSettings(sampling_ratio=sampling_ratio,
                                         numeric_to_dimension=numeric_to_dimension,
                                         dimension_distinct_cut_off=dimension_distinct_cut_off,
                                         use_dimension_columns=use_dimension_columns,
                                         rejected_dimension_columns=rejected_dimension_columns,
                                         use_quantity_columns=use_quantity_columns,
                                         rejected_quantity_columns=rejected_quantity_columns)

    def get_trx_descriptor(self, source: Union[Dict, DbObjectInput, pd.DataFrame, str] = None,
                           entity_column_name: str = None,
                           feature_gen_setting: Union[FeatureGenerationSettings, Dict] = None,
                           number_of_rows_per_step: int = None) -> TrxDescriptor:
        """
        Generates TrxDescriptor instance for AfeModelling.
        """
        raw_input_source = RecordSource(source=source) if source is not None else None
        reading_settings = AfeDataReadingSettings(number_of_rows_per_step=number_of_rows_per_step)
        if feature_gen_setting is not None:
            if isinstance(feature_gen_setting, list):
                if len(feature_gen_setting) > 1:
                    raise ValueError("Execution with multiple feature generation settings is "
                                     "available only in premium version.")
                feature_gen_setting = feature_gen_setting[0]
            if isinstance(feature_gen_setting, dict):
                feature_gen_setting = self.get_feature_generation_setting(**feature_gen_setting)

        trx_desc = TrxDescriptor(modelling_raw_input_source=raw_input_source,
                                 entity_column_name=entity_column_name,
                                 feature_gen_setting=feature_gen_setting,
                                 reading_settings=reading_settings)
        return trx_desc

    @staticmethod
    def get_target_column(target_column_name: str = None,
                          target_column_type: Union[str, AfeTargetColumnType] = None,
                          target_positive_category: str = None,
                          target_negative_category: str = None,
                          target_indeterminate_category: str = None,
                          target_exclusion_category: str = None):
        """Generates AfeTargetColumn instance"""
        if target_column_name is None:
            return None
        return AfeTargetColumn(column_name=target_column_name,
                               target_column_type=target_column_type,
                               binary_target_info=BinaryTarget(positive_category=target_positive_category,
                                                               negative_category=target_negative_category,
                                                               indeterminate_category=target_indeterminate_category,
                                                               exclusion_category=target_exclusion_category))

    def get_target_descriptor(
            self,
            entity_column_name: str,
            date_column_name: str = None,
            date_column_type: Union[str, AfeDateColumnType] = None,
            date_column_format: str = None,
            date_column_db_format: str = None,
            target_column_name: str = None,
            target_column_type: Union[str, AfeTargetColumnType] = None,
            target_positive_category: str = None,
            target_negative_category: str = None,
            target_indeterminate_category: str = None,
            target_exclusion_category: str = None,
            default_measurement_date: datetime = None,
            number_of_rows_per_step: int = None
    ) -> TargetDescriptor:
        """
        Generates TargetDescriptor instance for AfeModelling.
        """
        # pylint: disable=too-many-arguments
        date_column = self.get_date_column(date_column_name=date_column_name,
                                           date_column_type=date_column_type,
                                           date_column_format=date_column_format,
                                           date_column_db_format=date_column_db_format)
        target_column = self.get_target_column(target_column_name=target_column_name,
                                               target_column_type=target_column_type,
                                               target_positive_category=target_positive_category,
                                               target_negative_category=target_negative_category,
                                               target_indeterminate_category=target_indeterminate_category,
                                               target_exclusion_category=target_exclusion_category)
        reading_settings = AfeDataReadingSettings(number_of_rows_per_step=number_of_rows_per_step)
        return TargetDescriptor(entity_column_name=entity_column_name,
                                date_column=date_column,
                                target_column=target_column,
                                default_measurement_date=default_measurement_date,
                                reading_settings=reading_settings)

    @staticmethod
    def get_supervised_algorithm_settings(training_percentage: float = None,
                                          min_data_in_leaf_and_sample_size_control_ratio: float = None,
                                          dimension_compression_ratio: float = None,
                                          dimension_max_cardinality: int = None,
                                          reduction_coverage: float = None,
                                          model_params: Dict = None,
                                          model_fit_params: Dict = None,
                                          final_reduction_coverage: float = None,
                                          final_model_params: Dict = None,
                                          final_model_fit_params: Dict = None,
                                          ):
        """
        Generates AfeSupervisedAlgorithmSettings instance.
        """
        # pylint: disable=too-many-arguments
        alg_settings: AfeSupervisedAlgorithmSettings = AfeSupervisedAlgorithmSettings(
            training_percentage=training_percentage,
            min_data_in_leaf_and_sample_size_control_ratio=min_data_in_leaf_and_sample_size_control_ratio,
            dimension_compression_ratio=dimension_compression_ratio,
            dimension_max_cardinality=dimension_max_cardinality
        )
        model_settings = ModelSettings(model_params=model_params, model_fit_params=model_fit_params,
                                       reduction_coverage=reduction_coverage)
        final_model_settings = ModelSettings(model_params=final_model_params, model_fit_params=final_model_fit_params,
                                             reduction_coverage=final_reduction_coverage)
        alg_settings.model_settings = model_settings
        alg_settings.final_model_settings = final_model_settings
        return alg_settings

    @staticmethod
    def get_unsupervised_algorithm_settings(bin_count: int = None,
                                            r_factor: float = None,
                                            max_column_count: int = None,
                                            is_logging: bool = None,
                                            dimension_compression_ratio: float = None,
                                            dimension_max_cardinality: int = None
                                            ):
        """
        Returns AfeUnsupervisedAlgorithmSettings instance.
        """
        # pylint: disable=too-many-arguments
        return AfeUnsupervisedAlgorithmSettings(
            bin_count=bin_count,
            r_factor=r_factor,
            max_column_count=max_column_count,
            is_logging=is_logging,
            dimension_compression_ratio=dimension_compression_ratio,
            dimension_max_cardinality=dimension_max_cardinality
        )

    @staticmethod
    def get_output_settings(output_folder: str = None,
                            output_prefix: str = None,
                            feature_name_prefix: str = None,
                            enable_feature_lookup_output_to_csv: bool = None,
                            enable_write_output: bool = None,
                            return_all_afe_columns: bool = None,
                            enable_all_feature_lookup_output_to_csv: bool = None
                            ):
        """Generates AfeOutputSettings instance"""
        # pylint: disable=too-many-arguments
        return AfeOutputSettings(output_folder=output_folder, output_prefix=output_prefix,
                                 feature_name_prefix=feature_name_prefix,
                                 enable_feature_lookup_output_to_csv=enable_feature_lookup_output_to_csv,
                                 enable_write_output=enable_write_output,
                                 return_all_afe_columns=return_all_afe_columns,
                                 enable_all_feature_lookup_output_to_csv=enable_all_feature_lookup_output_to_csv
                                 )

    def get_settings(
            self,
            trx_descriptor: Union[TrxDescriptor, Dict] = None,
            target_descriptor: Union[TargetDescriptor, Dict] = None,
            target_record_source_list: List[Union[str, pd.DataFrame]] = None,
            output_settings: Union[AfeOutputSettings, Dict] = None,
            scoring_raw_input_source: Union[Dict, DbObjectInput, str, pd.DataFrame] = None,
            scoring_target_source: Union[Dict, DbObjectInput, str, pd.DataFrame] = None,
            afe_learning_type: Union[str, AfeLearningType] = None,
            algorithm_settings: AfeAlgorithmSettings = None,
            auto_column_decider_settings: AutoColumnDeciderSettings = None,
            number_of_cores: int = None,
            max_number_of_transaction_samples: int = None,
            max_number_of_target_samples: int = None,
            entity_table_existence: bool = None
    ) -> AfeModellingSettings:
        """
        Generates AfeModellingSettings instance.
        """
        # pylint: disable=too-many-arguments
        scoring_raw_input_source = RecordSource(
            source=scoring_raw_input_source) if scoring_raw_input_source is not None else None
        scoring_target_source = RecordSource(
            source=scoring_target_source) if scoring_target_source is not None else None
        if target_record_source_list is not None:
            target_record_source_list = [RecordSource(source=obj) for obj in target_record_source_list]

        process_settings = AfeProcessSettings(number_of_cores=number_of_cores)

        if trx_descriptor is not None and isinstance(trx_descriptor, dict):
            trx_descriptor = self.get_trx_descriptor(**trx_descriptor)

        if target_descriptor is not None:
            if isinstance(target_descriptor, dict):
                target_descriptor = self.get_target_descriptor(**target_descriptor)

        if output_settings is not None:
            if isinstance(output_settings, Dict):
                output_settings = self.get_output_settings(**output_settings)
        else:
            output_settings = AfeOutputSettings()

        modelling_settings = AfeModellingSettings()
        modelling_settings.data_source_settings = AfeDataSettings(
            trx_descriptor=trx_descriptor,
            target_descriptor=target_descriptor,
            target_record_source_list=target_record_source_list,
            auto_column_decider_settings=auto_column_decider_settings,
            max_number_of_target_samples=max_number_of_target_samples,
            max_number_of_transaction_samples=max_number_of_transaction_samples
        )

        modelling_settings.data_source_settings.entity_table_existence = entity_table_existence
        modelling_settings.scoring_raw_input_source = scoring_raw_input_source
        modelling_settings.scoring_target_source = scoring_target_source
        modelling_settings.algorithm_settings = algorithm_settings
        modelling_settings.process_settings = process_settings
        modelling_settings.output_settings = output_settings
        modelling_settings.afe_learning_type = afe_learning_type
        return modelling_settings

    @classmethod
    def _get_default_included_afe_operators(cls) -> List[AfeOperator]:
        return [AfeOperator.Density, AfeOperator.Sum, AfeOperator.Frequency, AfeOperator.TimeSinceFirst,
                AfeOperator.Ratio]



"""This module keeps the sweep enumerations."""
from enum import Enum, auto


class RegressionAttributeSelectionMethod(Enum):
    """Enum for regression attribute selection method"""
    STEPWISE = auto()
    FORWARD = auto()
    NONE = auto()


class AttributeSelectionStatus(Enum):
    """Enum for attribute selection status"""
    ENFORCED = auto()
    INCLUDED = auto()
    EXLUDED = auto()
    NEVER_INCLUDED = auto()


class TestStatisticsType(Enum):
    """Enum for test statistics type"""
    F_STATISTICS = auto()
    T_STATISTICS = auto()
    CHI_SQUARE = auto()



"""Includes DateValueDefinition class."""
from typing import List

from organon.fl.core.enums.datetime_token import DatetimeToken


class DateValueDefinition:
    """Date value definition class."""

    def __init__(self, year: int = None, month: int = None, day: int = None, hour: int = None):
        self.year: int = year
        self.month: int = month
        self.day: int = day
        self.hour: int = hour

    def get_datetime_tokens(self) -> List[DatetimeToken]:
        """Returns tokens for supplied values in date value definition"""
        tokens = []
        if self.year is not None:
            tokens.append(DatetimeToken.YEAR)
        if self.month is not None:
            tokens.append(DatetimeToken.MONTH)
        if self.day is not None:
            tokens.append(DatetimeToken.DAY)
        if self.hour is not None:
            tokens.append(DatetimeToken.HOUR)
        return tokens



"""
This module includes AfeColumnDto class.
"""


class AfeColumnDto:
    """
    AfeColumnDto
    """

    def __init__(self):
        self.name: str = None
        self.dimension: str = None
        self.dimension_set: str = None
        self.quantity: str = None
        self.operator: str = None
        self.date_column_name: str = None
        self.time_window: int = None
        self.trend_time_window: int = None
        self.time_resolution: str = None
        self.offset: int = None



"""This module includes helper functions and constants for parameter configurations."""
from typing import Dict

USE_CLASS_DEFAULT_STR = "NOT_SET"


def get_params(params: Dict) -> Dict:
    """filter parameters that wanted to be used as class default"""
    return {key: value for key, value in params.items() if value is not USE_CLASS_DEFAULT_STR}



"""This module includes AfeDateColumnSettings class."""
from datetime import datetime
from typing import List

from organon.afe.domain.settings.afe_date_column import AfeDateColumn
from organon.afe.domain.settings.temporal_grid import TemporalGrid
from organon.afe.domain.enums.date_resolution import DateResolution
from organon.afe.domain.enums.afe_operator import AfeOperator
from organon.afe.domain.common.reader_helper import get_values_from_kwargs


class FeatureGenerationSettings:
    """AFE settings per date column in transaction data"""

    ATTR_DICT = {
        "date_column": AfeDateColumn,
        "dimension_columns": List[str],
        "quantity_columns": List[str],
        "temporal_grids": List[TemporalGrid],
        "date_resolution": DateResolution,
        "horizon_list": List[int],
        "included_operators": List[AfeOperator],
        "date_offset": int,
        "max_observation_date": datetime
    }

    def __init__(self, **kwargs):
        self.date_column: AfeDateColumn = None
        self.dimension_columns: List[str] = None
        self.quantity_columns: List[str] = None
        self.temporal_grids: List[TemporalGrid] = None
        self.date_resolution: DateResolution = None
        self.horizon_list: List[int] = None
        self.included_operators: List[AfeOperator] = None
        self.date_offset: int = None
        self.max_observation_date: datetime = None

        get_values_from_kwargs(self, self.ATTR_DICT, kwargs)



"""Includes FlSortedDict class."""
from typing import TypeVar, Generic, Dict

from sortedcontainers import SortedDict as _SortedDict

KT = TypeVar('KT')
VT = TypeVar('VT')


class SortedDict(Generic[KT, VT]):
    """A sorted dict implementation using sortedcontainers.SortedDict"""

    def __init__(self, values: Dict[KT, VT] = None):
        values = {} if values is None else values
        self._dict = _SortedDict(values)

    def keys(self):
        """returns keysview instance"""
        return self._dict.keys()

    def values(self):
        """returns valuesview instance"""
        return self._dict.values()

    def items(self):
        """returns dict_items instance"""
        return self._dict.items()

    def popitem(self, index=-1):
        """pop item in dict"""
        return self._dict.popitem(index)

    def __getitem__(self, key: KT) -> VT:
        return self._dict[key]

    def __setitem__(self, key: KT, value: VT):
        self._dict[key] = value

    def __len__(self):
        return len(self._dict)

    def __iter__(self):
        return iter(self._dict)

    def __str__(self):
        return str(self._dict)

    def __repr__(self):
        return repr(self._dict)



"""Includes AbstractModellingService class."""
import abc
import os
from datetime import datetime
from typing import Dict, List, Optional, Tuple
from typing import TypeVar, Generic

import numpy as np
import pandas as pd

from organon.afe.dataaccess.services.i_sample_data_service import ISampleDataService
from organon.afe.dataaccess.services.sample_data_service import SampleDataService
from organon.afe.domain.common.base_execution_plan_service import BaseExecutionPlanService
from organon.afe.domain.common.persist_helper import PersistHelper
from organon.afe.domain.enums.afe_learning_type import AfeLearningType
from organon.afe.domain.enums.record_source_type import RecordSourceType
from organon.afe.domain.modelling.businessobjects.afe_column import AfeColumn
from organon.afe.domain.modelling.businessobjects.base_afe_feature import BaseAfeFeature
from organon.afe.domain.modelling.businessobjects.data_frame_builder import DataFrameBuilder
from organon.afe.domain.modelling.businessobjects.multi_target_entity_container import MultiTargetEntityContainer
from organon.afe.domain.modelling.businessobjects.target_file_record_collection import TargetFileRecordCollection
from organon.afe.domain.modelling.businessobjects.transaction_file_record_collection import \
    TransactionFileRecordCollection
from organon.afe.domain.modelling.businessobjects.transaction_file_stats import TransactionFileStats
from organon.afe.domain.modelling.helper import trx_reader_helper
from organon.afe.domain.modelling.supervised.afe_supervised_algorithm_settings import AfeSupervisedAlgorithmSettings
from organon.afe.domain.modelling.supervised.multi_target_entity_container_service import \
    MultiTargetEntityContainerService
from organon.afe.domain.modelling.unsupervised.afe_unsupervised_algorithm_settings import \
    AfeUnsupervisedAlgorithmSettings
from organon.afe.domain.reporting.afe_column_dto import AfeColumnDto
from organon.afe.domain.reporting.base_afe_model_output import BaseAfeModelOutput
from organon.afe.domain.reporting.runtime_statistics import RuntimeStatistics
from organon.afe.domain.settings.afe_algorithm_settings import AfeAlgorithmSettings
from organon.afe.domain.settings.base_afe_modelling_settings import BaseAfeModellingSettings
from organon.afe.domain.settings.feature_generation_settings import FeatureGenerationSettings
from organon.afe.domain.settings.modelling_reduction_settings import ModellingReductionSettings
from organon.afe.domain.settings.record_source import RecordSource
from organon.fl.core.businessobjects.date_interval import DateInterval
from organon.fl.core.businessobjects.dict_dataframe import DictDataFrame
from organon.fl.core.businessobjects.fl_core_static_objects import FlCoreStaticObjects
from organon.fl.core.businessobjects.idata_partition import IDataPartition
from organon.fl.core.executionutil.objects.stopwatch import StopWatch
from organon.fl.core.extensions.string_extensions import effective_value
from organon.fl.core.fileoperations import file_io_helper, directory_helper
from organon.fl.core.helpers import guid_helper, date_helper, process_info_helper
from organon.fl.logging.helpers.log_helper import LogHelper

AfeModellingSettingsType = TypeVar("AfeModellingSettingsType", bound=BaseAfeModellingSettings)
AfeModelOutputType = TypeVar("AfeModelOutputType", bound=BaseAfeModelOutput)
AfeColumnType = TypeVar("AfeColumnType", bound=AfeColumn)


class AbstractModellingService(Generic[AfeModellingSettingsType, AfeModelOutputType, AfeColumnType],
                               metaclass=abc.ABCMeta):
    """Base class for afe modelling."""

    def __init__(self, settings: AfeModellingSettingsType):
        self._settings: AfeModellingSettingsType = settings
        self._all_afe_columns: Optional[Dict[str, AfeColumnType]] = {}
        self._exec_plan_service = None

    def execute(self, num_of_threads: int = None) -> AfeModelOutputType:
        """Executes multi target afe modelling using given number of threads"""
        LogHelper.info("Afe ModellingService.Execute started")
        self._control_date_column_constraint()
        if num_of_threads is None:
            num_of_threads = self._settings.process_settings.number_of_cores

        self._exec_plan_service = self._get_exec_plan_service()

        self.__log_exec_plan()
        self._exec_plan_service.start_recording_runtime_stats(interval_in_seconds=1)
        try:
            watch = StopWatch(True)

            reduced_columns, reduction_extra_info_per_target, trx_record_collection = \
                self.__execute_until_reduction(num_of_threads)

            self._exec_plan_service.start_output()
            watch.restart()

            reduced_columns_list = list(reduced_columns.values())

            output_features = self._get_output_features(reduced_columns_list)
            all_features = self._get_output_features(list(self._all_afe_columns.values()))
            self._persist_lookup_data(output_features, trx_record_collection.transaction_file_stats)
            self._persist_all_lookup_data(all_features, trx_record_collection.transaction_file_stats)

            output_time = watch.get_elapsed_seconds(True)
            self._exec_plan_service.complete_output(output_time)

            self.__log_exec_plan()
            self._exec_plan_service.stop_recording_runtime_stats()

            runtime_stats = self._exec_plan_service.get_runtime_statistics()  # pylint:disable=assignment-from-none
            if runtime_stats is not None:
                LogHelper.info(f"CPU Time: {runtime_stats.user_cpu_time + runtime_stats.system_cpu_time} seconds")
            try:
                LogHelper.info(
                    f"Process peak memory usage: {process_info_helper.get_peak_memory_usage() / (1024 * 1024)} MB")
            except NotImplementedError:
                pass

            return self._get_model_output(output_features,
                                          all_features,
                                          trx_record_collection.transaction_file_stats,
                                          runtime_stats, reduction_extra_info_per_target)
        finally:
            self._exec_plan_service.stop_recording_runtime_stats()

    def _persist_lookup_data(self, output_features: Dict[str, BaseAfeFeature], stats: TransactionFileStats):
        output_settings = self._settings.output_settings
        write_csv = output_settings.enable_feature_lookup_output_to_csv and output_settings.enable_write_output
        if not write_csv:
            return
        lookup_data_table = self._features_to_data_frame(list(output_features.values()), stats)
        csv_data = self._feature_frame_to_csv(lookup_data_table)
        file_prefix = self._settings.output_settings.get_feature_lookup_table_name()
        self._persist_lookup_csv(file_prefix, csv_data)

    def _persist_all_lookup_data(self, all_features: Dict[str, BaseAfeFeature], stats: TransactionFileStats):
        output_settings = self._settings.output_settings
        write_csv = output_settings.enable_all_feature_lookup_output_to_csv and output_settings.enable_write_output
        if not write_csv:
            return
        lookup_data_table = self._features_to_data_frame(list(all_features.values()), stats)
        csv_data = self._feature_frame_to_csv(lookup_data_table)
        file_prefix = self._settings.output_settings.get_all_features_lookup_table_name()
        self._persist_lookup_csv(file_prefix, csv_data)

    @classmethod
    def _features_to_data_frame(cls, features: List[BaseAfeFeature], stats: TransactionFileStats) -> pd.DataFrame:
        return PersistHelper.get_lookup_data_frame(features, stats)

    def _get_exec_plan_service(self) -> BaseExecutionPlanService:
        # pylint: disable=no-self-use
        return BaseExecutionPlanService()

    def __log_exec_plan(self):
        plan = self._exec_plan_service.get_plan()  # pylint:disable=assignment-from-none
        if plan is not None:
            LogHelper.info(f"Plan is: {plan}")

    @abc.abstractmethod
    def _get_output_features(self, reduced_columns_list: List[AfeColumn]):
        """Generates afe feature instances from afe columns"""

    @abc.abstractmethod
    def _control_date_column_constraint(self):
        pass

    def __execute_until_reduction(self, num_of_threads: int) -> \
            Tuple[Dict[str, AfeColumn], dict, TransactionFileRecordCollection]:

        watch = StopWatch(True)
        LogHelper.info("Started populating multi-target-entity-container")
        self._exec_plan_service.start_get_targets()
        rejected_target_classes = None
        if self._settings.data_source_settings.target_descriptor.target_column is not None:
            binary_target_info = self._settings.data_source_settings. \
                target_descriptor.target_column.binary_target_info
            rejected_target_classes = [binary_target_info.exclusion_category,
                                       binary_target_info.indeterminate_category]
        multi_target_entity_container = self._get_multi_target_entity_container(rejected_target_classes)

        interval_per_entity_per_date_col = self._get_interval_per_entity_per_date_col(multi_target_entity_container)

        self._exec_plan_service.start_decide_columns()
        sample_service = self._get_sample_service(interval_per_entity_per_date_col)
        self._decide_dimension_and_quantity_columns(sample_service)

        decide_time = watch.get_elapsed_seconds(True)
        self._exec_plan_service.complete_decide_columns(decide_time)

        self._exec_plan_service.start_trx_file()
        data_source_settings = self._settings.data_source_settings
        trx_descriptor = data_source_settings.trx_descriptor
        all_d_cols, all_q_cols = trx_descriptor.get_all_dimension_and_quantity_columns()
        trx_file_record_collection = self._read_transactions_file(all_d_cols,
                                                                  all_q_cols,
                                                                  interval_per_entity_per_date_col,
                                                                  multi_target_entity_container.unique_entity_list)
        trx_time = watch.get_elapsed_seconds(True)
        self._exec_plan_service.complete_trx_file(trx_time)

        self._exec_plan_service.set_collections(trx_file_record_collection,
                                                multi_target_entity_container.records_per_file)

        self._exec_plan_service.calculate_reduce_time()
        self.__log_exec_plan()

        self._exec_plan_service.start_reduce()
        watch.restart()

        reduced_columns, reduction_extra_info_per_target = self.__reduce_columns(
            multi_target_entity_container.records_per_file,
            trx_file_record_collection, num_of_threads)
        reduce_time = watch.get_elapsed_seconds(True)
        self._exec_plan_service.complete_reduce(reduce_time)
        LogHelper.info(f"Reducing columns finished in {reduce_time} seconds.")
        self._exec_plan_service.save_statistics()

        return reduced_columns, reduction_extra_info_per_target, trx_file_record_collection

    def _get_multi_target_entity_container(self, rejected_target_classes) -> MultiTargetEntityContainer:
        watch = StopWatch(True)
        multi_target_entity_container_service = MultiTargetEntityContainerService(
            self._settings.data_source_settings.target_record_source_list,
            self._settings.data_source_settings.target_descriptor,
            rejected_target_classes=rejected_target_classes)
        multi_target_entity_container_service.execute()
        self.control_target_record_files(multi_target_entity_container_service)
        target_time = watch.get_elapsed_seconds(True)
        self._exec_plan_service.complete_get_targets(target_time)
        LogHelper.info(
            f"Multi-target-entity-container has been populated in {target_time} seconds")

        multi_target_entity_container = multi_target_entity_container_service.container

        data_source_settings = self._settings.data_source_settings
        multi_target_entity_container.unify_entity_list(data_source_settings.max_number_of_target_samples)

        return multi_target_entity_container

    @abc.abstractmethod
    def _decide_dimension_and_quantity_columns(self, sample_service: ISampleDataService):
        """decides dimension and quantity columns for feature generation settings"""

    def _get_sample_service(self, interval_per_entity_per_date_col: Dict[str, Dict[str, DateInterval]]):
        return SampleDataService(self._settings.data_source_settings, self._settings.algorithm_settings,
                                 interval_per_entity_per_date_column=interval_per_entity_per_date_col)

    def _get_model_output(self, output_features: Dict[str, BaseAfeFeature],
                          all_features: Dict[str, BaseAfeFeature],
                          transaction_file_stats: TransactionFileStats,
                          runtime_statistics: RuntimeStatistics,
                          reduction_extra_info_per_target: Dict[RecordSource, dict]) -> AfeModelOutputType:
        # pylint: disable=too-many-arguments
        model_output = self._get_model_output_instance()
        model_output.model_identifier = guid_helper.new_guid(32)
        model_output.build_date = datetime.now()
        model_output.output_features = output_features
        model_output.all_features = all_features
        model_output.transaction_file_stats = transaction_file_stats
        model_output.runtime_stats = runtime_statistics
        data_source_settings = self._settings.data_source_settings
        model_output.trx_entity_column = data_source_settings.trx_descriptor.entity_column_name
        model_output.trx_date_columns = self._get_non_empty_date_columns_dict()
        model_output.target_descriptor = data_source_settings.target_descriptor

        if reduction_extra_info_per_target is not None:
            final_col_metrics = []
            for target_source in self._settings.data_source_settings.target_record_source_list:
                if target_source in reduction_extra_info_per_target and \
                        "final_column_metrics" in reduction_extra_info_per_target[target_source]:
                    final_col_metrics.append(reduction_extra_info_per_target[target_source]["final_column_metrics"])
            model_output.final_column_metrics = final_col_metrics

        return model_output

    @classmethod
    def _get_model_output_instance(cls) -> AfeModelOutputType:
        return BaseAfeModelOutput()

    @abc.abstractmethod
    def _get_non_empty_date_columns_dict(self):
        """return trx date columns  which are not None as a dict"""

    @abc.abstractmethod
    def _get_interval_per_entity_per_date_col(self, multi_target_entity_container: MultiTargetEntityContainer):
        """Return interval_per_entity_per_date_col for trx read"""

    def __reduce_columns(self,
                         records_per_file: Dict[RecordSource, TargetFileRecordCollection],
                         trx_file_record_collection, num_of_threads):
        reduced_columns = {}
        reduction_extra_info_per_target = {}
        record_source_index = 0
        for record_source, target_file_record_collection in records_per_file.items():
            reduction_settings = ModellingReductionSettings()
            reduction_settings.target_record_source = record_source
            reduction_settings.target_record_source_index = record_source_index
            reduction_settings.target_file_record_collection = target_file_record_collection
            reduction_settings.trx_file_record_collection = trx_file_record_collection
            reduction_settings.num_threads = num_of_threads
            _dict, reduction_extra_info = self.__reduce(reduction_settings)
            reduction_extra_info_per_target[record_source] = reduction_extra_info
            for key, value in _dict.items():
                if key not in reduced_columns:
                    if AbstractModellingService._check_reduced_columns(reduced_columns, value):
                        value.source = self._get_afe_column_source(record_source, record_source_index)
                        reduced_columns[key] = value
            record_source_index += 1

        return reduced_columns, reduction_extra_info_per_target

    @staticmethod
    def _check_reduced_columns(reduced_columns, afe_column_to_be_checked):
        check_fields = ['dimension_name', 'operator', 'quantity_name', 'set', 'time_window',
                        'date_column', 'in_out', 'offset', 'date_resolution']
        for _, afe_column in reduced_columns.items():
            is_equals = True
            for field in check_fields:
                if getattr(afe_column_to_be_checked, field) != getattr(afe_column, field):
                    is_equals = False
                    break
            if is_equals:
                return False
        return True

    def __reduce(self, reduction_settings: ModellingReductionSettings) \
            -> Tuple[Dict[str, AfeColumn], dict]:

        frame_builder = self._get_data_frame_builder(reduction_settings.target_file_record_collection,
                                                     reduction_settings.trx_file_record_collection)

        all_columns_frame = DictDataFrame(reduction_settings.target_file_record_collection.sampled_count)

        file = self.__get_report_file(self._settings.output_settings, reduction_settings.target_record_source,
                                      reduction_settings.target_record_source_index)

        directory_helper.create(self._settings.output_settings.output_folder)

        partition = self._reduce_for_settings(frame_builder, all_columns_frame, reduction_settings, file)
        _dict, reduction_extra_info = self.__run_final_model(reduction_settings.num_threads, partition, frame_builder,
                                                             file,
                                                             reduction_settings.trx_file_record_collection,
                                                             all_columns_frame)
        if file is not None:
            file.flush()
            file.close()

        return _dict, reduction_extra_info

    def _reduce_for_settings(self, frame_builder, all_columns_frame, reduction_settings, file) -> IDataPartition:
        """Executes reduction for all date-dim-qty triplets."""
        raise NotImplementedError

    def _reduce_for_feature_gen_setting(self, frame_builder, order,
                                        feature_generation_settings: FeatureGenerationSettings,
                                        reduction_settings, all_columns_frame, partition, file):
        # pylint: disable=too-many-arguments
        watch = StopWatch(start=True)
        algorithm_settings: AfeAlgorithmSettings = self._settings.algorithm_settings

        if feature_generation_settings.date_column is not None:
            date_col_name = feature_generation_settings.date_column.column_name
            frame_builder.trx_file_record_collection.sort(date_col_name)
        else:
            date_col_name = "No_Date"
        dim_col_index = 0
        for dimension_column in feature_generation_settings.dimension_columns:
            q_col_index = 0
            for quantity_column in feature_generation_settings.quantity_columns:
                watch.restart()
                target_file_name = self._get_target_file_name(reduction_settings.target_record_source,
                                                              reduction_settings.target_record_source_index)
                LogHelper.info(f"Target File: {target_file_name}, "
                               f"Date Column: {date_col_name}, "
                               f"Dimension: {dimension_column}, "
                               f"Quantity: {quantity_column}")

                frame_builder.execute(reduction_settings.num_threads, dimension_column, quantity_column,
                                      feature_generation_settings)

                df_time = watch.get_elapsed_seconds()
                LogHelper.info(f"Building Data Frame finished in {df_time} seconds")

                if partition is None:
                    partition = self._get_partition(frame_builder, algorithm_settings)

                self._add_afe_columns_to_all_afe_columns(frame_builder.name_to_column,
                                                         reduction_settings.target_record_source,
                                                         reduction_settings.target_record_source_index)
                watch.restart()

                selected_cols = self._get_selected_cols(reduction_settings.num_threads,
                                                        partition,
                                                        algorithm_settings,
                                                        frame_builder,
                                                        frame_builder.frame)
                model_time = watch.get_elapsed_seconds()
                LogHelper.info(f"Building non-interaction model finished in {model_time} seconds")

                for col in selected_cols:
                    all_columns_frame.try_add(col, frame_builder.frame.get_value(col))

                if self._settings.output_settings.enable_write_output:
                    PersistHelper.print_report(selected_cols, file, dimension_column, quantity_column,
                                               frame_builder.name_to_column,
                                               reduction_settings.trx_file_record_collection.transaction_file_stats)

                self._persist_frame_output(frame_builder,
                                           date_col_name,
                                           dimension_column, quantity_column,
                                           order, dim_col_index, q_col_index)
                LogHelper.info(f"Remaining input columns: {len(all_columns_frame.get_column_names())}")
                self.__update_exec_service(reduction_settings,
                                           order,
                                           dimension_column,
                                           frame_builder,
                                           selected_cols,
                                           df_time,
                                           model_time)
                watch.restart()
                frame_builder.clean_frame()
                LogHelper.info(f"Cleaned frame in {watch.get_elapsed_seconds(True)} seconds. ")
                LogHelper.info(f"Iteration finished for Target File: {target_file_name}, "
                               f"Dimension: {dimension_column}, "
                               f"Quantity: {quantity_column}")
                q_col_index += 1
            dim_col_index += 1
        return partition

    @abc.abstractmethod
    def _get_data_frame_builder(self, target_file_record_collection: TargetFileRecordCollection,
                                trx_file_record_collection: TransactionFileRecordCollection):
        """Return DataFrameBuilder instance"""

    @staticmethod
    def _get_partition(frame_builder: DataFrameBuilder = None,
                       algorithm_settings: AfeAlgorithmSettings = None) -> Optional[IDataPartition]:
        # pylint: disable=unused-argument
        return None

    def _get_selected_cols(self, num_threads: int,
                           partition: IDataPartition = None,
                           algorithm_settings: AfeAlgorithmSettings = None,
                           frame_builder: DataFrameBuilder = None,
                           frame_with_all_columns: DictDataFrame = None,
                           is_final=False) -> List[str]:

        raise NotImplementedError

    def _get_final_columns_metrics(self, num_threads: int, frame_builder: DataFrameBuilder,
                                   frame_with_all_columns: DictDataFrame, reduced_columns: List[str],
                                   algorithm_settings: AfeSupervisedAlgorithmSettings) -> dict:
        """
        Generate extra metrics for final selected AfeColumns
        """
        # pylint: disable=no-self-use,unused-argument
        return None

    def control_target_record_files(self, multi_target_entity_container_service: MultiTargetEntityContainerService):
        """
        control sample size and min_data_in_leaf ratio.
        :param multi_target_entity_container_service:
        :return: true or false
        """

    def __run_final_model(self, num_threads: int,
                          partition: IDataPartition,
                          frame_builder: DataFrameBuilder,
                          file,
                          trx_file_record_collection: TransactionFileRecordCollection,
                          frame_with_all_columns: DictDataFrame = None) -> Tuple[Dict[str, AfeColumn], dict]:
        watch = StopWatch(start=True)
        watch.restart()
        LogHelper.info("Building final model")
        algorithm_settings = self._settings.algorithm_settings
        selected_cols = self._get_selected_cols(num_threads,
                                                partition,
                                                algorithm_settings,
                                                frame_builder,
                                                frame_with_all_columns,
                                                is_final=True)
        # pylint: disable=assignment-from-none
        final_col_metrics = self._get_final_columns_metrics(num_threads, frame_builder, frame_with_all_columns,
                                                            selected_cols, algorithm_settings)
        reduction_extra_info = {"final_column_metrics": final_col_metrics}
        _dict: Dict[str, AfeColumn] = {}
        for col in selected_cols:
            afe_column = frame_builder.name_to_column[col]
            _dict[afe_column.column_name] = afe_column

        if self._settings.output_settings.enable_write_output:
            PersistHelper.print_report(selected_cols, file, "", "", frame_builder.name_to_column,
                                       trx_file_record_collection.transaction_file_stats)
        LogHelper.info(f"Building final model finished in {watch.get_elapsed_seconds(True)} seconds")

        return _dict, reduction_extra_info

    def _add_afe_columns_to_all_afe_columns(self, name_to_column: Dict[str, AfeColumnType],
                                            target_record_source: RecordSource,
                                            target_record_source_index: int):
        if not self._settings.output_settings.return_all_afe_columns and not \
                self._settings.output_settings.enable_all_feature_lookup_output_to_csv:
            return
        source = self._get_afe_column_source(target_record_source, target_record_source_index)
        for name, afe_column in name_to_column.items():
            if name not in self._all_afe_columns:  # pylint: disable=unsupported-membership-test
                afe_column.source = source
                # pylint: disable=unsupported-assignment-operation
                self._all_afe_columns[name] = afe_column

    def _persist_frame_output(self, frame_builder,  # pylint: disable=too-many-arguments
                              date_col_name, dimension_column, quantity_column,
                              date_col_index, dim_col_index, q_col_index):
        """Persists current frame"""

    @staticmethod
    def _persist_lookup_csv(file_prefix: str, csv_data):
        folder_root = os.path.join(FlCoreStaticObjects.get_master_output_directory(), "AfeLookupCsvOutput")
        directory_helper.create(folder_root)
        date_str = date_helper.format_date(date_helper.now(), "%Y%m%d_%H%M%S_%f")[:-3]
        lookup_file_name = f"{file_prefix}_{date_str}.csv"
        file_io_helper.write_to_file(os.path.join(folder_root, lookup_file_name), csv_data)

    def _persist_lookup_data_table(self, lookup_data_table):
        """Persist lookup data table"""
        # pylint: disable=no-self-use

    @staticmethod
    def _feature_frame_to_csv(lookup_data_frame: pd.DataFrame):
        csv_data = "FEATURE,DIMENSION,DIMENSION_SET,QUANTITY,OPERATOR,DATE_COLUMN," \
                   "TIME_WINDOW,RESOLUTION,OFFSET\n"
        for _, row in lookup_data_frame.iterrows():
            record = AfeColumnDto()
            feature = None if pd.isnull(row["FEATURE"]) else str(row["FEATURE"])
            record.name = None if pd.isnull(row["FEATURE_EXTENDED_NAME"]) else str(row["FEATURE_EXTENDED_NAME"])
            record.date_column_name = None if pd.isnull(row["DATE_COLUMN_NAME"]) else str(row["DATE_COLUMN_NAME"])
            record.dimension = None if pd.isnull(row["DIMENSION"]) else str(row["DIMENSION"])
            record.dimension_set = None if pd.isnull(row["DIMENSION_SET"]) else str(row["DIMENSION_SET"])
            record.quantity = None if pd.isnull(row["QUANTITY"]) else str(row["QUANTITY"])
            record.operator = None if pd.isnull(row["OPERATOR"]) else str(row["OPERATOR"])
            record.time_window = None if pd.isnull(row["TIME_WINDOW"]) else row["TIME_WINDOW"]
            record.time_resolution = None if pd.isnull(row["RESOLUTION"]) else str(row["RESOLUTION"])
            record.offset = None if pd.isnull(row["DATE_OFFSET"]) else row["DATE_OFFSET"]

            csv_data += ",".join([effective_value(feature), effective_value(record.dimension),
                                  effective_value(record.dimension_set), record.quantity, record.operator,
                                  effective_value(record.date_column_name),
                                  str(record.time_window),
                                  effective_value(record.time_resolution), str(record.offset)])

            csv_data += "\n"
        return csv_data

    def _read_transactions_file(self, d_columns, q_columns,
                                interval_per_entity_per_date_col: Dict[str, Dict[str, DateInterval]],
                                unique_entity_list: List[str]):
        watch = StopWatch(True)

        LogHelper.info("Started reading transactions file")

        trx_reader = trx_reader_helper.get_trx_file_record_reader_for_modelling(
            d_columns,
            q_columns,
            self._settings.data_source_settings,
            self._settings.algorithm_settings,
            interval_per_entity_per_date_col,
            unique_entity_list
        )

        trx_file_record_collection: TransactionFileRecordCollection = trx_reader.read()
        LogHelper.info(f"Finished reading transactions file in {watch.get_elapsed_seconds(True)} seconds")

        return trx_file_record_collection

    def __update_exec_service(self,
                              reduction_settings: ModellingReductionSettings,
                              order: int,
                              dimension_column: str,
                              frame_builder: DataFrameBuilder,
                              selected_cols: List[str],
                              df_time: float,
                              model_time: float):
        # pylint: disable=too-many-arguments
        if isinstance(self._exec_plan_service, BaseExecutionPlanService):
            return
        algorithm_settings: AfeSupervisedAlgorithmSettings or AfeUnsupervisedAlgorithmSettings = \
            self._settings.algorithm_settings
        if self._exec_plan_service.iteration_count == 0:
            self._exec_plan_service.calculate_calibrated_reduce_time(reduction_settings.target_record_source,
                                                                     order,
                                                                     dimension_column,
                                                                     df_time,
                                                                     model_time)
            plan = self._exec_plan_service.get_plan()  # pylint: disable=assignment-from-none
            if plan is not None:
                LogHelper.info(f'Calibrated plan is : {plan}')

        trx_collection = reduction_settings.trx_file_record_collection

        self._exec_plan_service.add_df_builder_statistic(
            reduction_settings.target_file_record_collection.sampled_count,
            trx_collection.actual_record_count,
            len(np.unique(trx_collection.d_arrays[:trx_collection.actual_record_count,
                          trx_collection.d_map[dimension_column]])),
            order, len(frame_builder.frame.get_column_names()), df_time)

        data_frame = frame_builder.frame.data_frame
        if self._settings.afe_learning_type == AfeLearningType.Supervised:
            self._exec_plan_service.add_supervised_statistic(data_frame.shape[0],
                                                             data_frame.shape[1], model_time)
        else:
            self._exec_plan_service.add_unsupervised_statistic(data_frame.shape[1],
                                                               algorithm_settings.bin_count,
                                                               data_frame.shape[0],
                                                               algorithm_settings.r_factor,
                                                               len(selected_cols),
                                                               model_time)

        self._exec_plan_service.iteration_count += 1
        LogHelper.info(f'{self._exec_plan_service.iteration_count} of '
                       f'{self._exec_plan_service.total_reduction_loop} reduction steps are completed')

    @classmethod
    def _get_afe_column_source(cls, target_record_source: RecordSource, record_source_index: int):
        target_file_name = cls._get_target_file_name(target_record_source, record_source_index)
        if target_record_source.get_type() == RecordSourceType.DATABASE:
            source = f"{target_file_name} at {target_record_source.source.connection_name}"
        else:
            source = target_file_name
        return source

    @classmethod
    def _get_target_file_name(cls, target_record_source: RecordSource, record_source_index: int):
        source_type = target_record_source.get_type()
        if source_type == RecordSourceType.TEXT:
            return target_record_source.source
        return f"Target {record_source_index}"

    @staticmethod
    def __get_report_file(output_settings, target_record_source, record_source_index):

        if target_record_source.get_type() == RecordSourceType.DATABASE:
            report_file_name = f"{output_settings.output_prefix}_{target_record_source.source.table_name}_REPORT.txt"
        else:
            report_file_name = f"{output_settings.output_prefix}_TARGET{record_source_index}_REPORT.txt"

        report_file_path = os.path.join(output_settings.output_folder, report_file_name)

        directory_helper.create(output_settings.output_folder)
        file = None
        if output_settings.enable_write_output:
            file = open(report_file_path, "w+", encoding="utf8")  # pylint: disable=consider-using-with

        return file



"""
This module includes TransactionFileDescriptor class.
"""
import abc

from organon.afe.domain.common.reader_helper import get_values_from_kwargs
from organon.afe.domain.settings.afe_reading_settings import AfeDataReadingSettings
from organon.afe.domain.settings.record_source import RecordSource


class BaseTrxDescriptor(metaclass=abc.ABCMeta):
    """
    Class for information about transaction file or database table
    """
    ATTR_DICT = {
        "modelling_raw_input_source": RecordSource,
        "entity_column_name": str,
        "reading_settings": AfeDataReadingSettings
    }

    def __init__(self, **kwargs):
        self.modelling_raw_input_source: RecordSource = None
        self.entity_column_name: str = None
        self.reading_settings: AfeDataReadingSettings = None

        get_values_from_kwargs(self, self.ATTR_DICT, kwargs)

    @abc.abstractmethod
    def get_all_dimension_and_quantity_columns(self):
        """Returns sets of all dimension and quantity columns defined in date column settings"""



"""
This module keeps the helper functions for boolean type.
"""


def to_int(bool_input: bool) -> int:
    """
    Convert boolean to int.
    True => 1
    False => 0
    :param bool_input: input boolean
    :return: integer representation of the boolean
    """
    if bool_input:
        return 1
    return 0


def from_int(num: int) -> bool:
    """
    Converts int to boolean.
    1 => True
    0 => False
    :param num: input integer
    :return: boolean representation of the integer
    """
    if num == 1:
        return True
    return False



"""Includes DqCalculationTaskExecutor class."""
from typing import List

from organon.fl.core.businessobjects.dataframe import DataFrame
from organon.idq.dataaccess.record_source_reader import RecordSourceReader
from organon.idq.domain.businessobjects.dq_calculation_result import DqCalculationResult
from organon.idq.domain.businessobjects.dq_data_column_collection import DqDataColumnCollection
from organon.idq.domain.businessobjects.dq_data_source import DqDataSource
from organon.idq.domain.businessobjects.statistics.data_source_statistics import DataSourceStatistics
from organon.idq.domain.businessobjects.statistics.population_statistics import PopulationStatistics
from organon.idq.domain.businessobjects.statistics.sample_statistics import SampleStatistics
from organon.idq.domain.enums.dq_record_source_type import DqRecordSourceType
from organon.idq.domain.enums.dq_run_type import DqRunType
from organon.idq.domain.services.statistics.source_statistics.df_source_stats_service import DfSourceStatsService
from organon.idq.domain.services.statistics.source_statistics.dq_source_stats_base_service import \
    DqSourceStatsBaseService
from organon.idq.domain.services.statistics.source_statistics.file_source_stats_service import FileSourceStatsService
from organon.idq.domain.settings.abstractions.dq_base_calculation_parameters import DqBaseCalculationParameters
from organon.idq.domain.settings.calculation.dq_df_calculation_parameters import DqDfCalculationParameters
from organon.idq.domain.settings.calculation.dq_file_calculation_parameters import DqFileCalculationParameters
from organon.idq.domain.settings.dq_column_metadata import DqColumnMetadata


class DqCalculationTaskExecutor:
    """Executes dq calculations with given parameters"""

    def __init__(self, calculation_parameters: DqBaseCalculationParameters,
                 run_type: DqRunType, is_test_calculation: bool = False):
        self.calculation_parameters: DqBaseCalculationParameters = calculation_parameters
        self.stats_service: DqSourceStatsBaseService = self._get_stats_service()
        self.is_test_calculation = is_test_calculation
        self.run_type = run_type

    def execute(self) -> DqCalculationResult:
        """Starts dq calculations"""
        if self.calculation_parameters.is_existing_calculation:
            return self._execute_for_existing_calculation()
        return self._execute_for_normal_calculation()

    def _execute_for_normal_calculation(self) -> DqCalculationResult:
        result = DqCalculationResult()
        result.calculation_name = self.calculation_parameters.calculation_name
        result.data_source_stats = DataSourceStatistics()
        result.population_stats = PopulationStatistics()

        sample_data = self._get_sample_data()
        data_col_collection = self._get_data_column_collection(sample_data)
        self._update_params_with_column_collection(data_col_collection)
        sample_stats = self._compute_sample_stats(sample_data, data_col_collection)
        filtered_columns = self._filter_high_cardinality_columns(data_col_collection)
        population_nominal_stats = self._get_population_nominal_stats(filtered_columns, data_col_collection)
        columns_metadata_list = self._update_params_with_column_collection(data_col_collection)

        result.data_source_stats.data_column_collection = data_col_collection
        result.data_source_stats.row_count = sample_data.full_data_row_count
        result.sample_stats = sample_stats
        result.column_metadata_list = columns_metadata_list
        result.population_stats.nominal_statistics = population_nominal_stats

        if self.is_test_calculation:
            result.sample_data = DataFrame()
            result.sample_data.data_frame = sample_data.sampled_data

        return result

    def _execute_for_existing_calculation(self) -> DqCalculationResult:
        if self.is_test_calculation:
            if self.calculation_parameters.input_source_settings.source.get_type() != DqRecordSourceType.DATA_FRAME:
                sample_data = self._get_sample_data()
                self.calculation_parameters.existing_result.sample_data = DataFrame()
                self.calculation_parameters.existing_result.sample_data.data_frame = sample_data.sampled_data
        return self.calculation_parameters.existing_result

    def _get_data_column_collection(self, sample_data: DqDataSource) -> DqDataColumnCollection:
        columns_dq_metadata = self.calculation_parameters.column_dq_metadata_list
        return self.stats_service.get_data_column_collection(sample_data, columns_dq_metadata)

    def _get_sample_data(self) -> DqDataSource:
        record_source_reader = RecordSourceReader(self.calculation_parameters.input_source_settings)
        return record_source_reader.read()

    def _compute_sample_stats(self, sample_data: DqDataSource, data_column_collection: DqDataColumnCollection) \
            -> SampleStatistics:
        return self.stats_service.get_sample_stats(sample_data.sampled_data, data_column_collection)

    def _update_params_with_column_collection(self, data_column_collection: DqDataColumnCollection) \
            -> List[DqColumnMetadata]:
        return self.stats_service.update_calc_params(data_column_collection)

    def _filter_high_cardinality_columns(self, data_column_collection: DqDataColumnCollection) -> List[str]:
        return self.stats_service.filter_high_cardinality_columns(data_column_collection)

    def _get_population_nominal_stats(self, filtered_columns: List[str],
                                      data_column_collection: DqDataColumnCollection):
        return self.stats_service.get_population_nominal_statistics(filtered_columns, data_column_collection)

    def _get_stats_service(self):

        if isinstance(self.calculation_parameters, DqFileCalculationParameters):
            return FileSourceStatsService(self.calculation_parameters)
        if isinstance(self.calculation_parameters, DqDfCalculationParameters):
            return DfSourceStatsService(self.calculation_parameters)
        raise NotImplementedError
